// ============================================================================
//
// Unit:    ibeDALConnectionFB_IB
// Inhalt:  Formular
// Aufgabe: DAL Connection fr Firebird Datenbank mit InterBase Express Komponenten
//
//   (c) 2015 by Klaus Etscheid
// ============================================================================
unit ibeDALConnectionFB_IB;

interface

uses
  SysUtils, Classes, DB,
  IBX.IBDatabase, IBX.IBQuery,
  ibeDALConnection, ibeDALQuery, ibeDALConnectionFB;

type

TibeDALConnectionFB_IB = class(TibeDALConnectionFB)
private
	FIBDB: TIBDatabase;
  FKeepConnection: Boolean;
	Property IBDB: TIBDatabase Read FIBDB Write FIBDB;
protected
	procedure InternalCreate(); override;
	procedure InternalSetup(); override;
  procedure InternalDAOQueryInit(AdaoQuery: TibeDALQuery); override;
  function InternalDatabaseExists(const ADatabaseName: string): Boolean; override;
  procedure InternalDatabaseCreate(const ADatabaseName: string); override;
  procedure InternalDatabaseDrop(const ADatabaseName: string); override;
	procedure InternalTranStart(); override;
	procedure InternalTranCommit(); override;
	procedure InternalTranRollback(aOrigException: Exception = nil); override;
  function GetDAOClass: TibeDALQueryClass; override;
  function GetActualKeepConnection(): Boolean; override;
  procedure SetActualKeepConnection(AKeep: Boolean); override;
  function GetInitialDatabase(): string; override;
  function GetActiveDatabase(): string; override;
  procedure SetActiveDatabase(const ADatabase: string); override;
  function GetIsSupportedIentityAfterInsert(): Boolean; override;
public
	destructor Destroy(); override;
  procedure Open(); override;
  procedure Close(); override;
  procedure DatasetInit(a_ds: TDataSet); override;
end;

implementation

uses
  ibeDALQueryFB_IB;

procedure TibeDALConnectionFB_IB.InternalCreate(); 
begin
  inherited InternalCreate();
  FKeepConnection := False;

  FIBDB := TIBDatabase.Create(nil);
  FIBDB.LoginPrompt := False;
  FIBDB.DefaultTransaction := TIBTransaction.Create(nil);
end;

destructor TibeDALConnectionFB_IB.Destroy(); 
begin
  inherited Destroy();

  FIBDB.DefaultTransaction.Free();
  FreeAndNil(FIBDB);
end;

procedure TibeDALConnectionFB_IB.InternalSetup(); 
begin
  UserName := Parameters.Values['User_Name'];
  Password := Parameters.Values['Password'];
  
  // ggf. relative Pfadangabe zu absoluten Pfad expandieren
  Parameters.Values['Database'] := ExpandFileName(Parameters.Values['Database']);

  IBDB.DatabaseName := Parameters.Values['Database'];
  IBDB.SQLDialect := StrToInt(Parameters.Values['SQLDialect']);
  IBDB.Params.Values['User_Name'] := UserName;
  IBDB.Params.Values['Password'] := Password;

  //IBDB.Params.Values['UseUnicode'] := Parameters.Values['UseUnicode'];
end;

procedure TibeDALConnectionFB_IB.InternalDAOQueryInit(AdaoQuery: TibeDALQuery); 
begin
  try
    DatasetInit(AdaoQuery.DataSet);
  except
    on E: Exception do begin
      raise EibeDALConnectionNotSupported.Create(Self, 'Firebird IB DAOQueryInit: Ein DAO der Klasse ' + AdaoQuery.ClassName + ' kann nicht verarbeitet werden', E);
    end;
  end;
end;

function TibeDALConnectionFB_IB.InternalDatabaseExists(const ADatabaseName: string): Boolean;
begin
  Result := FileExists(ADatabaseName);
end;

procedure TibeDALConnectionFB_IB.InternalDatabaseCreate(const ADatabaseName: string); 
begin
  IBDB.Close();
  IBDB.Params.Clear();
  IBDB.DatabaseName := ADatabaseName;
  IBDB.Params.Add (' user ''' + UserName + ''' password ''' + Password + ''' ');
  IBDB.Params.Add (' page_size 4096 ');
  IBDB.Params.Add (' default character set win1251 ');
  
  IBDB.CreateDatabase();
end;

procedure TibeDALConnectionFB_IB.InternalDatabaseDrop(const ADatabaseName: string); 
begin
  if FileExists(ADatabaseName) then
  begin
    IBDB.Close();
    IBDB.DatabaseName := ADatabaseName;
    IBDB.Params.Clear();
    IBDB.Params.Add('user_name=' + UserName);
    IBDB.Params.Add('password=' + Password);
    IBDB.Open();
    IBDB.DropDatabase();
  end;
end;

procedure TibeDALConnectionFB_IB.InternalTranStart();
begin
  if IBDB.DefaultTransaction.Active then
  begin
    // derzeit ist eine Transaktion aktiv. Diese erst Committen
    IBDB.DefaultTransaction.Commit();
  end;

  IBDB.DefaultTransaction.StartTransaction();
end;

procedure TibeDALConnectionFB_IB.InternalTranCommit(); 
begin
  IBDB.DefaultTransaction.Commit();
  IBDB.DefaultTransaction.Active := False;
end;

procedure TibeDALConnectionFB_IB.InternalTranRollback(aOrigException: Exception = nil); 
begin
  IBDB.DefaultTransaction.Rollback();
  IBDB.DefaultTransaction.Active := False;
end;

function TibeDALConnectionFB_IB.GetDAOClass: TibeDALQueryClass; 
begin
  Result := TibeDALQueryFB_IB;
end;

function TibeDALConnectionFB_IB.GetActualKeepConnection(): Boolean; 
begin
  Result := FKeepConnection;
end;

procedure TibeDALConnectionFB_IB.SetActualKeepConnection(AKeep: Boolean); 
begin
  FKeepConnection := AKeep;
end;

procedure TibeDALConnectionFB_IB.Open(); 
begin
  IBDB.Open();
end;

procedure TibeDALConnectionFB_IB.Close(); 
begin
  IBDB.Close();
end;

procedure TibeDALConnectionFB_IB.DatasetInit(a_ds: TDataSet); 
begin
  inherited DatasetInit(a_ds);

  if not (a_ds is TIBQuery) then begin
    raise EibeDALConnectionNotSupported.Create(Self, 'IB DatasetInit: die Methode darf nur fr Datasets der Klasse ' + TIBQuery.ClassName + ' aufgerufen werden. bergeben wurde ein Dataset der Klasse ' + a_ds.ClassName + '.');
  end else begin
    TIBQuery(a_ds).Database := IBDB;
    TIBQuery(a_ds).Transaction := IBDB.DefaultTransaction;
  end;
end;

function TibeDALConnectionFB_IB.GetInitialDatabase(): string;
begin
  Result := Parameters.Values['Database'];
end;

function TibeDALConnectionFB_IB.GetActiveDatabase(): string; 
begin
  Result := IBDB.DatabaseName;
end;

procedure TibeDALConnectionFB_IB.SetActiveDatabase(const ADatabase: string);
begin
  IBDB.DatabaseName := ADatabase;
end;

function TibeDALConnectionFB_IB.GetIsSupportedIentityAfterInsert(): Boolean; 
begin
  Result := True;
end;

initialization

TibeDALConnectionFactory.RegisterConnetion(gc_sConnectionType_FB_IB, TibeDALConnectionFB_IB);

end.
