// ============================================================================
//
// Unit:    ibeDataObject
// Inhalt:  Formular
// Aufgabe: DataObject Basisklasse
//
//   (c) by Klaus Etscheid since 2015
// ============================================================================
unit ibeDataObject;

interface

uses
  SysUtils, Variants, Classes, DB,
  ibeString, ibeVariant,
  ibeDALConnection, ibeDALQuery, ibeMemTable,
  ibeDataObjectTypes, ibeDataObjectDefinition, ibeDataObjectFields;

type

TibeCustomDataObject = class
private
  FConnection: TibeDALConnection;
  FDODef: TibeDataObjectDefinition;
  FQuery: TibeDALQuery;
  FdatEdit: TibeMemTable;
  FDOClassName: string;
  FState: TibeDataObjectState;
  FFieldList: TibeDataObjectInternalFieldList;
  FPublicField: TibeDataObjectField;
  FTraceEnabled: Boolean;
  procedure InitDatasetFields(ADataSet: TDataSet);
  procedure InitFieldList();
  function GetRecordCount(): Integer;
  function GetFieldCount(): Integer;
  function GetField(const AFieldName: string): TibeDataObjectField;
  function GetFieldByIndex(AIndex: Integer): TibeDataObjectField;
  function GetPKName(): string;
  function GetPKValue(): Variant;
  procedure OnFieldValueChanged(AField: TibeDataObjectInternalField);
  function GetDefinitionFileName(): TFileName;
protected
  procedure InitQuery(AQuery: TibeDALQuery);
  procedure InitNewRecord(); virtual;
  procedure AppendToTrace(const AText: String; const AComment: string = '');
  procedure InternalFieldValueChanged(const AFieldName: string; AValue: Variant); virtual;
  function DoBeforePost(): Boolean; virtual;
  function DoAfterPost(APKValue: Variant): Boolean; virtual;
  procedure DoBeforeDelete(APKValue: Variant); virtual;
  procedure DoAfterDelete(APKValue: Variant); virtual;
  function RetrieveCustomPKValue(): Variant; virtual;
  property DOClassName: string read FDOClassName;
  property DODef: TibeDataObjectDefinition read FDODef;
  property Query: TibeDALQuery read FQuery;
  property FieldList: TibeDataObjectInternalFieldList read FFieldList;
  function GetDefinitionFilePath(): TFileName; virtual;
  property DefinitionFilePath: TFileName read GetDefinitionFilePath;
  property DefinitionFileName: TFileName read GetDefinitionFileName;
public
  constructor Create(AConnection: TibeDALConnection); virtual;
  destructor Destroy(); override;
  property Connection: TibeDALConnection read FConnection;
  procedure Refresh();
  procedure Insert();
  procedure Edit();
  function Post(): Boolean;
  procedure Delete(); overload;
  procedure Delete(APKValue: Variant); overload;
  ///  <summary> DeleteAll() lscht alle im DO geladenen Datenstze </summary>
  ///  <returns> Anzahl gelschter Datenstze </returns>
  function DeleteAll(): Integer;
  function First(): Boolean;
  function Next(): Boolean;
  function Prior(): Boolean;
  procedure Last();
  function BOF(): Boolean;
  function EOF(): Boolean;
  function Fetch(ANoExceptionOnCountMismatch: Boolean = False): Boolean; overload;
  function Fetch(APKValue: Variant; ANoExceptionOnCountMismatch: Boolean = False): Boolean; overload;
  procedure FilterClear();
  function FilterBuild(const AFieldName: string; AFilterValue: Variant): string; overload;
  function RetrieveSQLTextForFieldValue(const AFieldName: string; AValue: Variant; out AValueAsSQLText: string): Boolean;
  function FilterBuild(const AFieldName: string; AFilterValue: Variant; AOpeator: string): string; overload;
  procedure FilterAdd(const AFilter: string); overload;
  procedure FilterAdd(const AFieldName: string; AFilterValue: Variant); overload;
  procedure FilterAdd(const AFieldName: string; AFilterValue: Variant; AOpeator: string); overload;
  property PKName: string read GetPKName;
  property PKValue: Variant read GetPKValue;
  property State: TibeDataObjectState read FState;
  property RecordCount: Integer read GetRecordCount;
  property FieldCount: Integer read GetFieldCount;
  property Field[const AFieldName: string]: TibeDataObjectField read GetField;
  property FieldByIndex[AIndex: Integer]: TibeDataObjectField read GetFieldByIndex;
  procedure DataCopyToDataset(ADat: TibeMemTable);
  procedure DataCreateDatasetStructure(ADat: TibeMemTable);
  procedure DataCopyFieldValues(ADat: TibeMemTable);
  property TraceEnabled: Boolean read FTraceEnabled write FTraceEnabled;
end;

TibeDataObject = class(TibeCustomDataObject)
private
  procedure Init();
protected
public
  constructor Create(AConnection: TibeDALConnection); override;
end;

TibeDataObjectClass = class of TibeDataObject;

implementation

uses
  ibeAppServiceProviderDB;

constructor TibeDataObject.Create(AConnection: TibeDALConnection);
begin
  inherited Create(AConnection);

  Init();
end;

procedure TibeDataObject.Init();
begin
  DODef.InitDefinition(DOClassName, DefinitionFileName);

  InitQuery(Query);
end;

constructor TibeCustomDataObject.Create(AConnection: TibeDALConnection);
begin
  inherited Create();
  FDODef := TibeDataObjectDefinition.Create();
  FFieldList := TibeDataObjectInternalFieldList.Create();
  FPublicField := TibeDataObjectField.Create();
  FQuery := AConnection.DAOQueryCreate();
  FdatEdit := TibeMemTable.Create(nil);

  FState := dosNone;
  FConnection := AConnection;
  TraceEnabled := Connection.TraceEnabled;
  FDOClassName := Copy(ClassName, 2, Length(ClassName)); // vom Klassennamen das fhrende "T" abschneiden
end;

destructor TibeCustomDataObject.Destroy();
begin
  inherited Destroy();

  FreeAndNil(FDODef);
  FreeAndNil(FFieldList);
  FreeAndNil(FPublicField);
  FreeAndNil(FQuery);
  FreeAndNil(FdatEdit);
end;

procedure TibeCustomDataObject.InitDatasetFields(ADataSet: TDataSet);
var
  i: Integer;
  LViewField: TibeDODefViewField;
  LDBField: TField;
begin
  if ADataSet.Tag <> DODef.ActiveView.FieldModifyTag then
  begin
    // Mit dem genderten ModifyTag signalisiert die View dass sie gendert wurde.
    // Die Feldlisten von DataObject und Abfrage mssen in diesem Fall neu
    // aufgebaut werden.
    ADataSet.Fields.Clear();
    ADataSet.FieldDefs.Clear();
    //Query.DataSet.FieldDefs.Update();

    FieldList.Clear();
    for i := 0 to DODef.ActiveView.FieldCount - 1 do
    begin
      LViewField := DODef.ActiveView.Field[i];
      LDBField := LViewField.AddFieldToDataset(ADataSet);
      FieldList.AddField(LViewField, LDBField, OnFieldValueChanged);
    end;

    // Das ModifyTag der View fr nachfolgende Vergleiche merken
    ADataSet.Tag := DODef.ActiveView.FieldModifyTag;
  end;
end;

procedure TibeCustomDataObject.InitFieldList();
var
  i: Integer;
  LViewField: TibeDODefViewField;
begin
  if FieldList.DefFieldsModifiedTag <> DODef.ActiveView.FieldModifyTag then
  begin
    // Mit dem genderten ModifyTag signalisiert die View dass sie gendert wurde.
    // Die Feldliste vom DataObject muss in diesem Fall neu aufgebaut werden.
    FieldList.Clear();

    for i := 0 to DODef.ActiveView.FieldCount - 1 do
    begin
      LViewField := DODef.ActiveView.Field[i];
      FieldList.AddField(LViewField, nil, OnFieldValueChanged);
    end;

    // Das ModifyTag der View fr nachfolgende Vergleiche merken
    FieldList.DefFieldsModifiedTag := DODef.ActiveView.FieldModifyTag;
  end;
end;

procedure TibeCustomDataObject.InitQuery(AQuery: TibeDALQuery);
begin
  AQuery.Close();
  AQuery.SQL := DODef.SQLSelect;
  AppendToTrace(AQuery.SQL);

  //InitDatasetFields(AQuery.DataSet);
  InitFieldList();
end;

procedure TibeCustomDataObject.Refresh();
begin
  InitQuery(Query);

  Query.Open();
  FieldList.BindDataSetFields(Query.DataSet);

  //FFieldList.ExchangeDataSet(Query.DataSet);

  FState := dosBrowse;
  
  First();
end;

procedure TibeCustomDataObject.Insert();
begin
  if not Query.DataSet.Active then
  begin
    Refresh();
  end;

  FdatEdit.Close();
//  InitDatasetFields(FdatEdit);
//  FFieldList.ExchangeDataSet(FdatEdit);
  InitFieldList();
  FieldList.AddDBFieldsToDataset(FdatEdit);
  FieldList.BindDataSetFields(FdatEdit);
  FdatEdit.Open();
  FdatEdit.Append();

  FState := dosInsert;

  InitNewRecord();
end;

procedure TibeCustomDataObject.InitNewRecord();
begin
end;

procedure TibeCustomDataObject.Edit();
var
  LNativeValue: Variant;
  LTblField: TibeDODefTableField;
begin
  LTblField := DODef.FieldByName[PKName];
  if Assigned(LTblField) then
  begin
    LNativeValue := LTblField.ConvertToTableFieldValue(PKValue);

    if Query.DataSet.Locate(PKName, LNativeValue, []) then
    begin
      FdatEdit.Close();
  //    InitDatasetFields(FdatEdit);
  //    FFieldList.ExchangeDataSet(FdatEdit);
      InitFieldList();
      FieldList.AddDBFieldsToDataset(FdatEdit);
      FieldList.BindDataSetFields(FdatEdit);
      FdatEdit.LoadFromDataset(Query.DataSet, False);

      FdatEdit.Edit();

      FState := dosEdit;
    end else
    begin
      raise EibeDataObjectDefinition.Create('DataObject "' + DOClassName + '": Edit nicht mglich, der Datensatz ist nicht geladen');
    end;
  end else
  begin
    raise EibeDataObjectDefinition.Create('DataObject "' + DOClassName + '": Edit nicht mglich, das PK-Feld "' + PKName + '" ist in der Datenmenge nicht vorhanden');
  end;
end;

function TibeCustomDataObject.Post(): Boolean;
var
  i: Integer;
  LPKValue: Variant;
  LRecordPosted: Boolean;
  LValueAsSQLText: string;
begin
  Connection.TranStart();
  try
    DODef.ClearPostParams();

    LRecordPosted := DoBeforePost();
  
    for i := 0 to FFieldList.Count - 1 do
    begin
      if FFieldList[i].Modified                                           // nur genderte Felder aktualisieren
        and SameText(FFieldList[i].TblAlias, DODef.MainTableAlias)        // nur wenn das Feld aus der MainTable kommt
        and (not SameText(FFieldList[i].FieldName, DODef.MainTablePKName))  // das PK-Feld nur bei Insert verarbeiten
      then begin
        if not RetrieveSQLTextForFieldValue(FFieldList[i].FieldName, FFieldList[i].Value, LValueAsSQLText) then
        begin
          raise EibeDataObject.Create('DataObject "' + DOClassName + '": Parameter fr Feld "' + FFieldList[i].FieldName + '" kann nicht angelegt werden, das Feld wurde nicht gefunden');
        end else
        begin
          DODef.AddPostParam(FFieldList[i].FieldName, LValueAsSQLText);
        end;
      end;
    end;

    if (State = dosInsert) or (DODef.PostParamCount > 0) then
    begin
      // neuer Datensatz oder es wurden Werte gendert. Daten abschicken
      if State = dosInsert then
      begin
        if DODef.MainTableKKValueMode in [pkmCustom] then
        begin
          case DODef.MainTableKKValueMode of
            pkmCustom:  LPKValue := PKValue;
          end;

          if not VarHasValue(LPKValue) then
          begin
            raise EibeDataObject.Create('Datenobjekt ' + DOClassName + '.Insert(): fr das PK-Feld ' + PKName + ' ist kein Wert angegeben');
          end;

          if not RetrieveSQLTextForFieldValue(PKName, LPKValue, LValueAsSQLText) then
          begin
            raise EibeDataObject.Create('DataObject "' + DOClassName + '": Parameter fr Feld "' + PKName + '" kann nicht angelegt werden, das Feld wurde nicht gefunden');
          end else
          begin
            DODef.AddPostParam(PKName, LValueAsSQLText);
          end;
        end;

        Query.SQL := DODef.SQLInsert;
      end else if State = dosEdit then
      begin
        LPKValue := PKValue;
        RetrieveSQLTextForFieldValue(PKName, LPKValue, LValueAsSQLText);
        FilterClear();
        DODef.PKFilterAdd(LValueAsSQLText, True);
        Query.SQL := DODef.SQLUpdate;
      end;

      AppendToTrace(Query.SQL);

      try
        Query.Execute();
        LRecordPosted := True;
        if State = dosInsert then
        begin
          if DODef.MainTableKKValueMode = pkmSerial then
          begin
            LPKValue := Connection.RetrieveIdentityValueAfterInsert(DODef.MainTablePKSequenceName);
          end;
        end;

        if LRecordPosted then
        begin
          LRecordPosted := DoAfterPost(LPKValue);
        end;

        // den neuen/genderten Datensatz neu laden
        Fetch(LPKValue);
      finally
        FState := dosNone;
      end;
    end;

    Connection.TranCommit();
  except
    on E: Exception do
    begin
      Connection.TranRollback(E);
      raise;
    end;
  end;

  Result := LRecordPosted;
end;

procedure TibeCustomDataObject.Delete(APKValue: Variant);
var
  LValueAsSQLText: string;
begin
  Connection.TranStart();
  try
    Fetch(APKValue);

    DoBeforeDelete(APKValue);

    RetrieveSQLTextForFieldValue(PKName, APKValue, LValueAsSQLText);
    FilterClear();
    DODef.PKFilterAdd(LValueAsSQLText, True);
    Query.SQL := DODef.SQLDelete;

    AppendToTrace(Query.SQL);
  
    Query.Execute();

    DoAfterDelete(APKValue);
    
    Connection.TranCommit();
  except
    on E: Exception do
    begin
      Connection.TranRollback(E);
      raise;
    end;
  end;
end;

function TibeCustomDataObject.DeleteAll(): Integer;
var
  i: Integer;
  LPKArray: Variant;
//  LDat: TibeMemTable;
begin
  Result := 0;
  // Variantes Array zur Aufnahme der selektierten PKValues erstellen
  LPKArray := VarArrayCreate([0, Pred(RecordCount)], varVariant);

  First();
  i := 0;
  // PKValues aller geladenen Datenstze in das Array eintragen
  while not EOF do
  begin
    LPKArray[i] := PKValue;
    Inc(i);
    Next();
  end;

  Connection.TranStart();
  try
    // ber alle Arrayelemente iterieren und mit den hinterlegten PKValues die
    // Datenstze lschen
    for i := 0 to VarArrayHighBound(LPKArray, 1) do
    begin
      Delete(LPKArray[i]);
      Inc(Result);
    end;
    
    Connection.TranCommit();
  except
    on E: Exception do
    begin
      Connection.TranRollback(E);
      raise;
    end;
  end;

//  Result := 0;
//  LDat := TibeMemTable.Create(nil);
//  try
//    LDat.LoadFromDataset(Query.DataSet, True);
//
//    LCtrlTran := Connection.TranStart();
//    try
//      // ber alle Arrayelemente iterieren und mit den hinterlegten PKValues die
//      // Datenstze lschen
//      LDat.First();
//      while not LDat.Eof do
//      begin
//        Delete(LDat.FieldByName(PKName).Value);
//        Inc(Result);
//        LDat.Next();
//      end;
//    
//      Connection.TranCommit();
//    except
//      on E: Exception do
//      begin
//        Connection.TranRollback(E);
//        raise;
//      end;
//    end;
//  finally
//    FreeAndNil(LDat);
//  end;
end;

procedure TibeCustomDataObject.Delete();
begin
  Delete(PKValue);
end;

function TibeCustomDataObject.First(): Boolean;
begin
  Query.DataSet.First();

  Result := not Query.DataSet.Eof;

  if Result then
  begin
    FieldList.InitFieldsFromDataset();
  end;
end;

function TibeCustomDataObject.Next(): Boolean;
begin
  Query.DataSet.Next();

  Result := not Query.DataSet.Eof;

  if Result then
  begin
    FieldList.InitFieldsFromDataset();
  end;
end;

function TibeCustomDataObject.Prior(): Boolean;
begin
  Query.DataSet.Prior();

  Result := not Query.DataSet.Bof;

  if Result then
  begin
    FieldList.InitFieldsFromDataset();
  end;
end;

procedure TibeCustomDataObject.Last();
begin
  Query.DataSet.Last();

  FieldList.InitFieldsFromDataset();
end;

function TibeCustomDataObject.BOF(): Boolean;
begin
  Result := Query.DataSet.Bof;
end;

function TibeCustomDataObject.EOF(): Boolean;
begin
  Result := Query.DataSet.Eof;
end;

function TibeCustomDataObject.Fetch(ANoExceptionOnCountMismatch: Boolean): Boolean;
begin
  Refresh();

  Result := RecordCount = 1;
  if not (Result or ANoExceptionOnCountMismatch) then
  begin
    raise EibeDOFetchCountMismatch.Create('Fehler in Datenobjekt ' + DOClassName + ': Fetch() muss genau einen Datensatz liefern, geladen wurden ' + IntToStr(RecordCount));
  end;
end;

function TibeCustomDataObject.Fetch(APKValue: Variant; ANoExceptionOnCountMismatch: Boolean): Boolean;
begin
  FilterClear();
  FilterAdd(PKName, APKValue);
  Refresh();

  Result := RecordCount = 1;
  if not (Result or ANoExceptionOnCountMismatch) then
  begin
    raise EibeDOFetchCountMismatch.Create('Fehler in Datenobjekt ' + DOClassName + ': Fetch() muss genau einen Datensatz liefern, geladen wurden ' + IntToStr(RecordCount));
  end;
end;

procedure TibeCustomDataObject.FilterClear();
begin
  DODef.FilterClear();
end;

function TibeCustomDataObject.FilterBuild(const AFieldName: string; AFilterValue: Variant): string;
begin
  Result := FilterBuild(AFieldName, AFilterValue, '=');
end;

function TibeCustomDataObject.RetrieveSQLTextForFieldValue(const AFieldName: string; AValue: Variant; out AValueAsSQLText: string): Boolean;
var
  LNativeValue: Variant;
  //LInternalField: TibeDataObjectInternalField;
  LTblField: TibeDODefTableField;
begin
  Result := False;
//  LInternalField := FFieldList.FieldByName[AFieldName];
//
//  if Assigned(LInternalField) then
//  begin
//    // das Feld ist in der Feldliste vom DO bekannt. Damit den nativen Feldwert bilden
//    LNativeValue := LInternalField.ConvertToTableFieldValue(AValue);
//    Result := True;
//  end else
//  begin
//    // das Feld ist in der DO-Feldliste nicht enthalten. Auf die Tabellenfelder
//    // in der DO-Definition ausweichen
//    LTblField := DODef.FieldByName[AFieldName];
//    if Assigned(LTblField) then
//    begin
//      LNativeValue := LTblField.ConvertToTableFieldValue(AValue);
//      Result := True;
//    end;
//  end;

  LTblField := DODef.FieldByName[AFieldName];
  if Assigned(LTblField) then
  begin
    LNativeValue := LTblField.ConvertToTableFieldValue(AValue);
    Result := True;
  end;

  AValueAsSQLText := Connection.CVToString(LNativeValue);
end;

function TibeCustomDataObject.FilterBuild(const AFieldName: string; AFilterValue: Variant; AOpeator: string): string;
var
  LValueAsSQLText: string;
  LValueString: Variant;
begin
  if AFieldName = '' then
  begin
    raise EibeDataObject.Create('Datenobjekt ' + DOClassName + ': FilterBuild() kann mit leerem Feldnamen nicht aufgerufen werden');
  end else
  begin
    if VarHasValue(AFilterValue) then
    begin
      if not RetrieveSQLTextForFieldValue(AFieldName, AFilterValue, LValueAsSQLText) then
      begin
        raise EibeDataObject.Create('Datenobjekt ' + DOClassName + ': FilterBuild() fr Feld "' + AFieldName + '" nicht mglich, das Feld ist in der View + "' + DODef.ActiveView.Name + '" nicht bekannt');
      end else
      begin
        LValueString := LValueAsSQLText;
      end;
    end else
    begin
      LValueString := Null;
    end;
  end;
  
  Result := DODef.FilterBuild(AFieldName, LValueString, AOpeator);
end;

procedure TibeCustomDataObject.FilterAdd(const AFilter: string);
begin
  DODef.FilterAdd(AFilter);
end;

procedure TibeCustomDataObject.FilterAdd(const AFieldName: string; AFilterValue: Variant);
begin
  FilterAdd(AFieldName, AFilterValue, '=');
end;

procedure TibeCustomDataObject.FilterAdd(const AFieldName: string; AFilterValue: Variant; AOpeator: string);
begin
  FilterAdd(FilterBuild(AFieldName, AFilterValue, AOpeator));
end;

procedure TibeCustomDataObject.OnFieldValueChanged(AField: TibeDataObjectInternalField);
begin
  InternalFieldValueChanged(AField.FieldName, AField.Value);
end;

procedure TibeCustomDataObject.InternalFieldValueChanged(const AFieldName: string; AValue: Variant);
begin
end;

function TibeCustomDataObject.DoBeforePost(): Boolean;
begin
  Result := True;
end;

function TibeCustomDataObject.DoAfterPost(APKValue: Variant): Boolean;
begin
  Result := True;
end;

procedure TibeCustomDataObject.DoBeforeDelete(APKValue: Variant);
begin
end;

procedure TibeCustomDataObject.DoAfterDelete(APKValue: Variant);
begin
end;

function TibeCustomDataObject.RetrieveCustomPKValue(): Variant;
begin
  Result := Field[PKName].Value;
end;

function TibeCustomDataObject.GetDefinitionFilePath(): TFileName;
begin
  Result := AddBackslash(RetrieveASP.CfgPath) + 'DOD\';
end;

function TibeCustomDataObject.GetDefinitionFileName(): TFileName;
var
  LDefinitionFileFound: Boolean;
  LDefnitionFileName: TFileName;
  LDOClassName: string;
  LDOFileClass: TibeDataObjectClass;
begin
  LDefinitionFileFound := False;
  LDefnitionFileName := '';
  LDOFileClass := TibeDataObjectClass(ClassType);
  LDOClassName := Copy(ClassName, 2, Length(ClassName)); // vom Klassennamen das fhrende "T" abschneiden

  while not LDefinitionFileFound and (LDOFileClass <> TibeCustomDataObject) do
  begin
    LDefnitionFileName := AddBackslash(DefinitionFilePath) + LDOClassName + '.dod';
    LDefinitionFileFound := FileExists(LDefnitionFileName);
    LDOFileClass := TibeDataObjectClass(LDOFileClass.ClassParent());
    LDOClassName := Copy(LDOFileClass.ClassName, 2, Length(LDOFileClass.ClassName)); // vom Klassennamen das fhrende "T" abschneiden
  end;

  if not LDefinitionFileFound then
  begin
    raise EibeDataObjectDefinition.Create(Self, 'Definition fr Datenobjekt ' + ClassName + ' kann nicht initialisiert werden, es wurde keine Definitionsdatei gefunden.');
  end;

  Result := LDefnitionFileName;
end;

function TibeCustomDataObject.GetRecordCount(): Integer;
//var
//  LBookmark: TBookmark;
begin
  Result := Query.RecordCount;
//  LBookmark := Query.DataSet.GetBookmark();
//  try
//    Query.DataSet.Last();
//    Result := Query.DataSet.RecordCount;
//    Query.DataSet.GotoBookmark(LBookmark);
//  finally
//    Query.DataSet.FreeBookmark(LBookmark);
//  end;
end;

function TibeCustomDataObject.GetFieldCount(): Integer;
begin
  Result := FFieldList.Count;
end;

function TibeCustomDataObject.GetField(const AFieldName: string): TibeDataObjectField;
begin
  if AFieldName = '' then
  begin
    raise EibeDataObject.Create('DataObject "' + DOClassName + '": auf Feld mit leerem Feldnamen kann nicht zugegriffen werden');
  end else
  begin
    FPublicField.SetInternalField(FieldList.FieldByName[AFieldName]);

    if not FPublicField.IsInternalFieldAssigned then
    begin
      raise EibeDataObject.Create('DataObject "' + DOClassName + '": Feld mit Feldnamen "' + AFieldName + '" ist nicht vorhanden');
    end;

    Result := FPublicField;
  end;
end;

function TibeCustomDataObject.GetFieldByIndex(AIndex: Integer): TibeDataObjectField;
begin
  FPublicField.SetInternalField(FieldList[AIndex]);

  Result := FPublicField;
end;

function TibeCustomDataObject.GetPKName(): string;
begin
  //Result := DODef.MainTablePKName;
  Result := DODef.ActiveView.ViewPKName;
end;

function TibeCustomDataObject.GetPKValue(): Variant;
var
  LPKName: string;
  LField: TibeDataObjectInternalField;
begin
  //LPKName := DODef.MainTablePKName;
  LPKName := DODef.ActiveView.ViewPKName;
  LField := FFieldList.FieldByName[LPKName];
  if not Assigned(LField) then
  begin
    raise EibeDataObject.Create('DataObject "' + DOClassName + '": PK-Feld mit Feldnamen "' + LPKName + '" ist nicht vorhanden');
  end else
  begin
    //Result := LField.ConvertToTableFieldValue(LField.Value);
    Result := LField.Value;
  end;
end;

procedure TibeCustomDataObject.DataCopyToDataset(ADat: TibeMemTable);
var
  LField: TibeDataObjectInternalField;
begin
  ADat.Close();
  DODef.ActiveView.CreateDatasetStructure(ADat);
  ADat.Open();

  First();
  while not EOF() do
  begin
    ADat.Append();
    for LField in FieldList do
    begin
      ADat.Fields[LField.DBFieldIndex].Value := LField.Value;
    end;
    ADat.Post();
    Next();
  end;
end;

procedure TibeCustomDataObject.DataCreateDatasetStructure(ADat: TibeMemTable);
begin
  ADat.Close();
  DODef.ActiveView.CreateDatasetStructure(ADat);
  ADat.Open();
end;

procedure TibeCustomDataObject.DataCopyFieldValues(ADat: TibeMemTable);
var
  LField: TibeDataObjectInternalField;
begin
  for LField in FieldList do
  begin
    ADat.Fields[LField.DBFieldIndex].Value := LField.Value;
  end;
end;

procedure TibeCustomDataObject.AppendToTrace(const AText: String; const AComment: string = '');
var
  LText: string;
  LFileMode: Word;
  LStream: TFileStream;
begin
  try
    if (RetrieveASP.DOTraceEnabled or TraceEnabled) and (Length(AText) > 0) then
    begin
      if not FileExists(RetrieveASP.DOTraceFileName) then
      begin
        LFileMode := fmCreate or fmShareDenyWrite;
      end else
      begin
        LFileMode := fmOpenWrite or fmShareDenyWrite;
      end;
      
      LStream := TFileStream.Create(RetrieveASP.DOTraceFileName, LFileMode);
      try
        LStream.Position := LStream.Size;
        LText := #13#10                                                      // Zeilenumbruch
               + '-- '                                                       // SQL Kommentar einleiten
               + FormatDateTime('yyyy-mm-dd hh:nn:ss.zzz', Now()) + ' '      // Timestamp
               + DOClassName                                                 // Klassenname vom DO
               + '(0x' + IntToHex(Integer(Self), 2 * SizeOf(Integer)) + ')'  // Adresse der DO-Instanz
               + #13#10;                                                     // Zeilenumbruch
        LStream.Write(LText[1], Length(LText) * SizeOf(LText[1]));

        if AComment <> '' then
        begin
          LText := '-- ' + AComment;
          LStream.Write(LText[1], Length(LText) * SizeOf(LText[1]));
        end;
        
        LStream.Write(AText[1], Length(AText) * SizeOf(AText[1]));
      finally
        FreeAndNil(LStream);
      end;
    end;
  except
    // nix machen
  end;
end;

end.
