// ============================================================================
//
// Unit:    ibeDataObjectDefinition
// Inhalt:  Formular
// Aufgabe: Einlesen der DataObject Definitionsdatei und bereitstellen der
//          Definitionsklassen
//
//   (c) by Klaus Etscheid since 2015
// ============================================================================
unit ibeDataObjectDefinition;

interface

uses
  SysUtils, StrUtils, System.Classes, Math, Generics.Defaults, Generics.Collections,
  XMLDoc, XMLIntf, DB, Variants,
  ibeString, ibeVariant, ibeFuncDateUtil, ibeXMLDocument, ibeDataObjectTypes;

type

TibeDODefTableList = class;
TibeDODefTableFieldList = class;

TibeDODefTable = class
private
  FXMElement: TibeXMLElement;
  FName, FAlias, FFieldPrefix, FJoinType, FJoinCondition, FDescription: string;
protected
public
  constructor Create();
  destructor Destroy(); override;
  procedure LoadFromXMLNode(ANode: IXMLNode);
  procedure SaveToXMLNode(ANode: IXMLNode);
  procedure Assign(ADefTable: TibeDODefTable);
  property Name: string read FName write FName;
  property Alias: string read FAlias write FAlias;
  property FieldPrefix: string read FFieldPrefix write FFieldPrefix;
  property JoinType: string read FJoinType write FJoinType;
  property JoinCondition: string read FJoinCondition write FJoinCondition;
  property Description: string read FDescription write FDescription;
end;

TibeDODefTableList = class(TObjectList<TibeDODefTable>)
private
protected
public
  function AddTable(): TibeDODefTable;
  function GetByAlias(const ATblAlias: string): TibeDODefTable;
  procedure LoadFromXMLNode(ANode: IXMLNode);
end;

TibeDODefTableField = class
private
  FTable: TibeDODefTable;
  FTblAlias, FName: string;
  FAlias, FExpression, FCaption, FDisplayOption: string;
  FWidth: Integer;
  FDataType: TibeDODefDataType;
  FXMElement: TibeXMLElement;
  function GetCaption(): string;
  function GetDBFieldClass(): TFieldClass;
  function GetQualifiedTblFieldName(): string;
  function GetIsExpression(): Boolean;
protected
  property XMElement: TibeXMLElement read FXMElement;
  property Table: TibeDODefTable read FTable;
public
  constructor Create(); overload;
  destructor Destroy(); override;
  function ConvertToTableFieldValue(AValue: Variant): Variant;
  function ConvertFromTableFieldValue(AValue: Variant): Variant;
  procedure LoadFromXMLNode(ANode: IXMLNode; ATableList: TibeDODefTableList);
  procedure SaveToXMLNode(ANode: IXMLNode);
  procedure Assign(ADefTableField: TibeDODefTableField);
  property TblAlias: string read FTblAlias write FTblAlias;
  property Name: string read FName write FName;
  property QualifiedTblFieldName: string read GetQualifiedTblFieldName;
  property Alias: string read FAlias write FAlias;
  property Expression: string read FExpression write FExpression;
  property Caption: string read GetCaption write FCaption;
  property Width: Integer read FWidth write FWidth;
  property DisplayOption: string read FDisplayOption write FDisplayOption;
  property DataType: TibeDODefDataType read FDataType write FDataType;
  property DBFieldClass: TFieldClass read GetDBFieldClass;
  property IsExpression: Boolean read GetIsExpression;
end;

TibeDODefTableFieldList = class(TObjectList<TibeDODefTableField>)
private
protected
public
  function AddField(): TibeDODefTableField;
  function GetByFieldAlias(const AFieldAlias: string): TibeDODefTableField;
  procedure LoadFromXMLNode(ANode: IXMLNode; ATableList: TibeDODefTableList);
end;

TibeDODefViewField = class
private
  FTblField: TibeDODefTableField;
  FXMElement: TibeXMLElement;
  FTblFieldAlias, FFieldName, FCaption: string;
  FDataType: TibeDODefDataType;
  FAggFunc: TibeDODefAggFunc;
  FColumnWidth: Integer;
  FVisibleInList: Boolean;
  function GetTblFieldClass(): TFieldClass;
  function GetTblDataType(): TibeDODefDataType;
  function GetTblFieldCaption(): string;
  function GetTblAlias(): string;
  function GetFieldClass(): TFieldClass;
  function GetDataType(): TibeDODefDataType;
  function GetWidth(): Integer;
  function GetDisplayOption(): string;
  function GetSQLExpressionText: string;
  function GetSQLColumnText: string;
  procedure SetVisibleInList(AVisible: Boolean);
  procedure InitFieldAttributesFromTblField(ATblField: TibeDODefTableField);
protected
  property TblField: TibeDODefTableField read FTblField;
public
  constructor Create(); overload;
  constructor Create(ATblField: TibeDODefTableField; const ACaption: string = ''; AAggFunc: TibeDODefAggFunc = doafNone); overload;
  constructor CreateCountAll(const ATblAlias, AFieldName: string; ACaption: string = '');
  destructor Destroy(); override;
  procedure InitFieldAttributes(const ATblFieldAlias: string; AAggFunc: TibeDODefAggFunc; AFieldList: TibeDODefTableFieldList);
  procedure LoadFromXMLNode(ANode: IXMLNode; AFieldList: TibeDODefTableFieldList);
  procedure SaveToXMLNode(ANode: IXMLNode);
  procedure Assign(ADefViewField: TibeDODefViewField; ATblFieldList: TibeDODefTableFieldList);
  function AddFieldToDataset(ADataset: TDataSet): TField;
  property TblAlias: string read GetTblAlias;
  property TblFieldAlias: string read FTblFieldAlias;
  property TblFieldClass: TFieldClass read GetTblFieldClass;
  property TblFieldCaption: string read GetTblFieldCaption;
  property TblDataType: TibeDODefDataType read GetTblDataType;
  property FieldClass: TFieldClass read GetFieldClass;
  property FieldName: string read FFieldName write FFieldName;
  property DataType: TibeDODefDataType read GetDataType write FDataType;
  property Width: Integer read GetWidth;
  property DisplayOption: string read GetDisplayOption;
  property AggFunc: TibeDODefAggFunc read FAggFunc;
  property Caption: string read FCaption write FCaption;
  property ColumnWidth: Integer read FColumnWidth write FColumnWidth;
  property VisibleInList: Boolean read FVisibleInList write SetVisibleInList;
  property SQLExpressionText: string read GetSQLExpressionText;
  property SQLColumnText: string read GetSQLColumnText;
end;

TibeDODefViewFieldList = class(TObjectList<TibeDODefViewField>)
private
  function GetFieldByName(const AFieldName: string): TibeDODefViewField;
protected
public
  function AddField(): TibeDODefViewField;
  procedure LoadFromXMLNode(ANode: IXMLNode; AFieldList: TibeDODefTableFieldList);
  procedure SaveToXMLNode(ANode: IXMLNode);
  property FieldByName[const AFieldName: string]: TibeDODefViewField read GetFieldByName;
end;

TibeDODefFieldCommon = class
private
  FXMElement: TibeXMLElement;
protected
  property XMElement: TibeXMLElement read FXMElement;
public
  constructor Create(); virtual;
  destructor Destroy(); override;
  procedure LoadFromXMLNode(ANode: IXMLNode; AFieldList: TibeDODefViewFieldList); virtual; abstract;
end;

TibeDODefFieldListCommon<T: TibeDODefFieldCommon, constructor> = class(TObjectList<T>)
private
  FFieldNodeName: string;
protected
  property FieldNodeName: string read FFieldNodeName write FFieldNodeName;
public
  constructor Create();
  function AddField(): T;
  procedure LoadFromXMLNode(ANode: IXMLNode; AFieldList: TibeDODefViewFieldList);
end;

TibeDODefViewOrderByField = class(TibeDODefFieldCommon)
private
  FTblFieldAlias: string;
  FDirection: TibeOrderByDirection;
  FViewField: TibeDODefViewField;
  function GetSQLText(): string;
protected
public
  procedure LoadFromXMLNode(ANode: IXMLNode; AFieldList: TibeDODefViewFieldList); override;
  procedure SaveToXMLNode(ANode: IXMLNode);
  procedure Assign(AOrderByField: TibeDODefViewOrderByField);
  procedure InitFieldAttributes(const ATblFieldAlias: string; AFieldList: TibeDODefViewFieldList; ADirection: TibeOrderByDirection = obdAscending);
  property TblFieldAlias: string read FTblFieldAlias;
  property Direction: TibeOrderByDirection read FDirection;
  property SQLText: string read GetSQLText;
end;

TibeDODefViewOrderByFieldList = class(TibeDODefFieldListCommon<TibeDODefViewOrderByField>)
private
protected
public
  procedure SaveToXMLNode(ANode: IXMLNode);
end;

TibeDODefViewGroupByField = class(TibeDODefFieldCommon)
private
  FTblFieldAlias: string;
  FViewField: TibeDODefViewField;
  function GetSQLText(): string;
protected
public
  procedure LoadFromXMLNode(ANode: IXMLNode; AFieldList: TibeDODefViewFieldList); override;
  procedure SaveToXMLNode(ANode: IXMLNode);
  procedure Assign(AGroupByField: TibeDODefViewGroupByField);
  procedure InitFieldAttributes(const ATblFieldAlias: string; AFieldList: TibeDODefViewFieldList);
  property TblFieldAlias: string read FTblFieldAlias;
  property ViewField: TibeDODefViewField read FViewField;
  property SQLText: string read GetSQLText;
end;

TibeDODefViewGroupByFieldList = class(TibeDODefFieldListCommon<TibeDODefViewGroupByField>)
private
protected
public
  procedure SaveToXMLNode(ANode: IXMLNode);
end;

TibeDODefView = class
private
  FXMElement: TibeXMLElement;
  FFieldList: TibeDODefViewFieldList;
  FGroupByFieldList: TibeDODefViewGroupByFieldList;
  FOrderByFieldList: TibeDODefViewOrderByFieldList;
  FName, FDescription: string;
  FFieldModifyTag: Integer;
  FViewPKName: string;
  procedure SetName(const AName: string);
  function GetFieldCount(): Integer;
  function GetField(AIndex: Integer): TibeDODefViewField;
  function GetFieldByName(const AFieldName: string): TibeDODefViewField;
  function GetOrderByFieldCount(): Integer;
  function GetOrderByField(AIndex: Integer): TibeDODefViewOrderByField;
  function GetGroupByFieldCount(): Integer;
  function GetGroupByField(AIndex: Integer): TibeDODefViewGroupByField;
  procedure TouchFieldModifyTag();
  procedure OnFieldListChanged(Sender: TObject; const Item: TibeDODefViewField; Action: TCollectionNotification);
protected
public
  constructor Create(); overload;
  constructor Create(const AName: string); overload;
  destructor Destroy(); override;
  procedure LoadFromXMLNode(ANode: IXMLNode; AFieldList: TibeDODefTableFieldList);
  procedure SaveToXMLNode(ANode: IXMLNode);
  procedure CreateDatasetStructure(ADat: TDataSet);
  procedure AddField(const ATblFieldAlias: string; ATblFieldList: TibeDODefTableFieldList; const ACaption: string = ''; AAggFunc: TibeDODefAggFunc = doafNone); overload;
  procedure AddField(AViewField: TibeDODefViewField; ATblFieldList: TibeDODefTableFieldList); overload;
  procedure AddOrderByField(const ATblFieldAlias: string; ADirection: TibeOrderByDirection);
  procedure AddGroupByField(const ATblFieldAlias: string);
  property Name: string read FName write SetName;
  property Description: string read FDescription write FDescription;
  property ViewPKName: string read FViewPKName write FViewPKName;
  property FieldCount: Integer read GetFieldCount;
  property Field[AIndex: Integer]: TibeDODefViewField read GetField;
  property FieldByName[const AFieldName: string]: TibeDODefViewField read GetFieldByName;
  property OrderByFieldCount: Integer read GetOrderByFieldCount;
  property OrderByField[AIndex: Integer]: TibeDODefViewOrderByField read GetOrderByField;
  property GroupByFieldCount: Integer read GetGroupByFieldCount;
  property GroupByField[AIndex: Integer]: TibeDODefViewGroupByField read GetGroupByField;
  /// FieldModifyTag bleibt konstant, so lange die View unverndert ist
  property FieldModifyTag: Integer read FFieldModifyTag;
end;

TibeDODefViewList = class(TObjectList<TibeDODefView>)
private
protected
public
  function GetByName(const AViewName: string): TibeDODefView;
  function AddView(): TibeDODefView;
  procedure LoadFromXMLNode(ANode: IXMLNode; AFieldList: TibeDODefTableFieldList);
end;

TibeDODefFilter = class
private
  FFilter: string;
protected
public
  constructor Create(const AFilter: string);
  property Filter: string read FFilter write FFilter;
end;

TibeDODefFilterList = class(TObjectList<TibeDODefFilter>)
private
protected
public
  function Add(const AFilter: string): Integer;
end;

TibeDODefPostParameter = class
private
  FTblField: TibeDODefTableField;
  FValue: Variant;
  function GetFieldName(): string;
  function GetTblFieldName(): string;
protected
public
  constructor Create(ATblField: TibeDODefTableField; AValue: Variant);
  Property FieldName: string read GetFieldName;
  property TblFieldName: string read GetTblFieldName;
  property Value: Variant read FValue write FValue;
end;

TibeDODefPostParameterList = class(TObjectList<TibeDODefPostParameter>)
private
protected
public
end;

TibeDataObjectDefinition = class
private
  FSQLText, FSQLColumnDelimiter, FDOClassName, FMainTableAlias, FMainTablePKName: string;
  FMainTableKKValueMode: TibeDOPKValueMode;
  FMainTablePKSequenceName: string;
  FMainViewName: string;
  FDefinitionFilePath: TFileName;
  FActiveView: TibeDODefView;
  FTableList: TibeDODefTableList;
  FFieldList: TibeDODefTableFieldList;
  FViewList: TibeDODefViewList;
  FFilterList: TibeDODefFilterList;
  FPostParamList: TibeDODefPostParameterList;
  function GetSQLSelect(): string;
  function GetSQLFilter(): string;
  function GetSQLInsert(): string;
  function GetSQLUpdate(): string;
  function GetSQLDelete(): string;
  function GetMainTableName: string;
  function GetTableCount: Integer;
  function GetTable(AIndex: Integer): TibeDODefTable;
  function GetFieldCount: Integer;
  function GetField(AIndex: Integer): TibeDODefTableField;
  function GetFieldByName(const AFieldName: string): TibeDODefTableField;
  function GetViewCount: Integer;
  function GetView(AIndex: Integer): TibeDODefView;
  function GetViewByName(const AName: string): TibeDODefView;
  function GetPostParamCount(): Integer;
  procedure SetDefinitionFilePath(const APath: TFileName);
protected
  property FilterList: TibeDODefFilterList read FFilterList;
public
  constructor Create();
  destructor Destroy(); override;
  procedure InitDefinition(const ADOClassName: string; const ADefinitionFileName: TFileName); overload;
  procedure InitDefinition(const ADOClassName: string; AStream: TStream); overload;
  procedure SaveToFile(const ADefinitionFileName: TFileName);
  procedure SaveToStream(AStream: TStream);
  procedure FilterClear();
  function FilterBuild(const AFieldName: string; AFilterValue: Variant): string; overload;
  function FilterBuild(const AFieldName: string; AFilterValue: Variant; AOpeator: string): string; overload;
  procedure FilterAdd(const AFilter: string);
  procedure PKFilterAdd(const APKValue: string; AOmitTblAlias: Boolean);
  property TableCount: Integer read GetTableCount;
  property Table[AIndex: Integer]: TibeDODefTable read GetTable;
  function AddTable(): Integer;
  property FieldCount: Integer read GetFieldCount;
  property Field[AIndex: Integer]: TibeDODefTableField read GetField;
  property FieldByName[const AFieldName: string]: TibeDODefTableField read GetFieldByName;
  function AddField(): Integer;
  property ViewCount: Integer read GetViewCount;
  property View[AIndex: Integer]: TibeDODefView read GetView;
  property ViewByName[const AName: string]: TibeDODefView read GetViewByName;
  procedure SelectActiveView(const AViewName: string);
  function AddView(const AName: string; AActivate: Boolean = True): Integer;
  procedure AddViewField(const AFieldName: string; const ACaption: string = ''; AAggFunc: TibeDODefAggFunc = doafNone); overload;
  procedure AddViewField(const AViewField: TibeDODefViewField); overload;
  procedure AddViewOrderByField(const AFieldName: string; ADirection: TibeOrderByDirection);
  procedure AddViewGroupByField(const AFieldName: string);
  procedure AddPostParam(const AFieldName: string; AValue: Variant);
  property PostParamCount: Integer read GetPostParamCount;
  procedure ClearPostParams();
  property MainTableAlias: string read FMainTableAlias write FMainTableAlias;
  property MainTablePKName: string read FMainTablePKName write FMainTablePKName;
  property MainTableName: string read GetMainTableName;
  property MainTableKKValueMode: TibeDOPKValueMode read FMainTableKKValueMode write FMainTableKKValueMode;
  property MainTablePKSequenceName: string read FMainTablePKSequenceName write FMainTablePKSequenceName;
  property MainViewName: string read FMainViewName write FMainViewName;
  property SQLSelect: string read GetSQLSelect;
  property SQLInsert: string read GetSQLInsert;
  property SQLUpdate: string read GetSQLUpdate;
  property SQLDelete: string read GetSQLDelete;
  property ActiveView: TibeDODefView read FActiveView;
  property DOClassName: string read FDOClassName;
  property DefinitionFilePath: TFileName read FDefinitionFilePath write SetDefinitionFilePath;
  property SQLColumnDelimiter: string read FSQLColumnDelimiter write FSQLColumnDelimiter;
end;

EibeDataObjectDefinition = class(EibeDataObject);

implementation

uses
  ibeAppServiceProvider;

// ===================== [ TibeDataObjectDefinition ] ==========================

constructor TibeDataObjectDefinition.Create();
begin
  inherited Create();
  FTableList := TibeDODefTableList.Create();
  FFieldList := TibeDODefTableFieldList.Create();
  FViewList := TibeDODefViewList.Create();
  FFilterList := TibeDODefFilterList.Create();
  FPostParamList := TibeDODefPostParameterList.Create();

  DefinitionFilePath := AddBackslash(RetrieveASP().CfgPath) + 'DOD\';
  //SQLColumnDelimiter := ', ';
  SQLColumnDelimiter := ',' + #13#10 + '  ';
  FMainViewName := 'Main';
end;

destructor TibeDataObjectDefinition.Destroy();
begin
  inherited Destroy();

  FreeAndNil(FPostParamList);
  FreeAndNil(FFilterList);
  FreeAndNil(FTableList);
  FreeAndNil(FFieldList);
  FreeAndNil(FViewList);
end;

function TibeDataObjectDefinition.GetSQLFilter(): string;
var
  i: Integer;
  LSQLFilter: string;
  LFilter: TibeDODefFilter;
begin
  i := 0;
  LSQLFilter := '';
  for LFilter in FilterList do
  begin
    if i > 0 then LSQLFilter := LSQLFilter + #13#10 + '  and ';
    LSQLFilter := LSQLFilter + '(' + LFilter.Filter + ')';
    Inc(i);
  end;

  Result := LSQLFilter;
end;

function TibeDataObjectDefinition.GetSQLSelect(): string;
var
  i: Integer;
  LSQL, LOrderByFields, LGroupByFields: string;
begin
  LSQL := 'select ';
  for i := 0 to ActiveView.FieldCount - 1 do
  begin
    if i > 0 then LSQL := LSQL + SQLColumnDelimiter;
    
    LSQL := LSQL + ActiveView.Field[i].SQLColumnText;
  end;

  LSQL :=LSQL + #13#10 + 'from ';
  for i := 0 to TableCount - 1 do
  begin    
    if i > 0 then LSQL := LSQL + #13#10 + '  ' + Table[i].JoinType + ' ';

    LSQL := LSQL + Table[i].Name + ' ' + Table[i].Alias;

    if i > 0 then LSQL := LSQL + ' on ' + Table[i].JoinCondition;
  end;

  if FilterList.Count > 0 then
  begin
    LSQL := LSQL + #13#10 + 'where ' + GetSQLFilter();
  end;

  if ActiveView.GroupByFieldCount > 0 then
  begin
    LGroupByFields := '';
    for i := 0 to ActiveView.GroupByFieldCount - 1 do
    begin
      if i > 0 then LGroupByFields := LGroupByFields + ', ';
      //LGroupByFields := LGroupByFields + ActiveView.GroupByField[i].TblAlias + '.' + ActiveView.GroupByField[i].Name;
      LGroupByFields := LGroupByFields + ActiveView.GroupByField[i].SQLText;
    end;
    LSQL := LSQL + #13#10 + 'group by ' + LGroupByFields;
  end;

  if ActiveView.OrderByFieldCount > 0 then
  begin
    LOrderByFields := '';
    for i := 0 to ActiveView.OrderByFieldCount - 1 do
    begin
      if i > 0 then LOrderByFields := LOrderByFields + ', ';
      LOrderByFields := LOrderByFields + ActiveView.OrderByField[i].SQLText;
    end;
    LSQL := LSQL + #13#10 + 'order by ' + LOrderByFields;
  end;
  
  Result := LSQL;
end;

function TibeDataObjectDefinition.GetSQLInsert(): string;
var
  LParam: TibeDODefPostParameter;
  LSQLColumns, LSQLValues: string;
begin
  if FPostParamList.Count = 0 then
  begin
    raise EibeDataObjectDefinition.Create('DataObject "' + DOClassName + '": Insert Statement kann nicht erstellt werden, die Paarameterliste ist leer');
  end else
  begin
    LSQLColumns := '';
    LSQLValues := '';
    for LParam in FPostParamList do 
    begin
      if LSQLColumns <> '' then
      begin
        LSQLColumns := LSQLColumns + ', ';
        LSQLValues := LSQLValues + ', ';
      end;
      LSQLColumns := LSQLColumns + LParam.TblFieldName;
      LSQLValues := LSQLValues + LParam.Value;
    end;

    Result := 'insert into ' + MainTableName + ' (' + LSQLColumns + ')' + #13#10 
            + 'values (' + LSQLValues + ')';
  end;
end;

function TibeDataObjectDefinition.GetSQLUpdate(): string;
var
  LParam: TibeDODefPostParameter;
  LSQLFields: string;
begin
  if FPostParamList.Count = 0 then
  begin
    raise EibeDataObjectDefinition.Create(Self, 'DataObject "' + DOClassName + '": Update Statement kann nicht erstellt werden, die Paarameterliste ist leer');
  end else
  begin
    if FilterList.Count = 0 then
    begin
      raise EibeDataObjectDefinition.Create(Self, 'DataObject "' + DOClassName + '": Update Statement kann nicht erstellt werden, es ist kein Filter definitier');
    end else
    begin
      LSQLFields := '';
      for LParam in FPostParamList do 
      begin
        if LSQLFields <> '' then
        begin
          LSQLFields := LSQLFields + ',';
        end;
        LSQLFields := LSQLFields + #13#10 + '  ' + LParam.TblFieldName + ' = ' + LParam.Value;
      end;

      Result := 'update ' + MainTableName + ' set' + LSQLFields + #13#10 + 'where ' + GetSQLFilter();
    end;
  end;
end;

function TibeDataObjectDefinition.GetSQLDelete(): string;
begin
  Result := 'delete from ' + MainTableName + ' where ' + GetSQLFilter();
end;

function TibeDataObjectDefinition.GetMainTableName: string;
var
  LTable: TibeDODefTable;
begin
  LTable := FTableList.GetByAlias(MainTableAlias);

  if Assigned(LTable) then
  begin
    Result := LTable.Name;
  end else
  begin
    raise EibeDataObjectDefinition.Create('MainTable in Datenobjekt ' + DOClassName + ' mit Alias ' + MainTableAlias + ' wurde nicht gefunden');
  end;
end;

procedure TibeDataObjectDefinition.InitDefinition(const ADOClassName: string; const ADefinitionFileName: TFileName);
var
  LStream: TFileStream;
begin

  if not FileExists(ADefinitionFileName) then
  begin
    raise EibeDataObjectDefinition.Create(Self, 'Definition fr Datenobjekt ' + ADOClassName + ' kann nicht initialisiert werden. Definitionsdatei "' + ADefinitionFileName + '" existiert nicht.');
  end;

  LStream := TFileStream.Create(ADefinitionFileName, fmOpenRead);
  try
    InitDefinition(ADOClassName, LStream);
  finally
    FreeAndNil(LStream);
  end;
end;

procedure TibeDataObjectDefinition.InitDefinition(const ADOClassName: string; AStream: TStream);
var
  LXMLDoc: TibeXMLDocument;
  LElement: TibeXMLElement;
  LNode: IXMLNode;
  LMainTableKKValueMode: string;
begin
  FDOClassName := ADOClassName; // DOClassName wird bei der Ausgabe von Fehlermeldungen verwendet

  LXMLDoc := TibeXMLDocument.Create();
  try
    LXMLDoc.LoadFromStream(AStream);

    LElement := TibeXMLElement.Create();
    try
      LElement.CurrentNode := LXMLDoc.DocumentElement;

      FMainTableAlias := LElement.GetAttributeTypedValue('MainTableAlias', '');
      FMainTablePKName := LElement.GetAttributeTypedValue('MainTablePKName', '');
      LMainTableKKValueMode := LElement.GetAttributeTypedValue('MainTablePKValueMode', PKValueModeArray[pkmCustom]);
      FMainTableKKValueMode := PKValueModeTypeFromString(LMainTableKKValueMode);
      FMainTablePKSequenceName := LElement.GetAttributeTypedValue('MainTablePKSequenceName', '');
      FMainViewName := LElement.GetAttributeTypedValue('MainViewName', FMainViewName);

      LNode := LElement.GetNodeByName('Tables');
      FTableList.LoadFromXMLNode(LNode);

      LNode := LElement.GetNodeByName('Fields');
      FFieldList.LoadFromXMLNode(LNode, FTableList);

      LNode := LElement.GetNodeByName('Views');
      FViewList.LoadFromXMLNode(LNode, FFieldList);

      SelectActiveView(FMainViewName);

      FSQLText := LElement.GetAttributeTypedValue('SQL', '');
    finally
      FreeAndNil(LElement);
    end;
  finally
    FreeAndNil(LXMLDoc);
  end;
end;

procedure TibeDataObjectDefinition.SaveToFile(const ADefinitionFileName: TFileName);
var
  LOpenMode: Integer;
  LStream: TFileStream;
begin
  if FileExists(ADefinitionFileName) then
  begin
    LOpenMode := fmOpenWrite;
  end else
  begin
    LOpenMode := fmCreate;
  end;
//  LOpenMode := fmCreate;
//  if FileExists(ADefinitionFileName) then
//  begin
//    DeleteFile(ADefinitionFileName);
//  end;

  LStream := TFileStream.Create(ADefinitionFileName, LOpenMode or fmShareDenyWrite);
  try
    LStream.Size := 0;
    SaveToStream(LStream);
  finally
    FreeAndNil(LStream);
  end;
end;

procedure TibeDataObjectDefinition.SaveToStream(AStream: TStream);
var
  i: Integer;
  LXMLDoc: TibeXMLDocument;
  LElement: TibeXMLElement;
  LParentNode, LChildNode: IXMLNode;
begin
  LXMLDoc := TibeXMLDocument.Create();
  try
    LXMLDoc.CreateDocument('DataObject');

    LElement := TibeXMLElement.Create();
    try
      LElement.CurrentNode := LXMLDoc.DocumentElement;
      LElement.CreateAttribute('MainTableAlias', MainTableAlias);
      LElement.CreateAttribute('MainTablePKName', MainTablePKName);
      LElement.CreateAttribute('MainTablePKValueMode', PKValueModeArray[MainTableKKValueMode]);
      LElement.CreateAttribute('MainTablePKSequenceName', MainTablePKSequenceName);
      LElement.CreateAttribute('MainViewName', FMainViewName);

      LParentNode := LElement.CreateElement('Tables');
      for i := 0 to TableCount - 1 do
      begin
        LChildNode := LElement.CreateElement('Table', LParentNode);

        Table[i].SaveToXMLNode(LChildNode);
      end;

      LParentNode := LElement.CreateElement('Fields');
      for i := 0 to FieldCount - 1 do
      begin
        LChildNode := LElement.CreateElement('Field', LParentNode);

        Field[i].SaveToXMLNode(LChildNode);
      end;

      LParentNode := LElement.CreateElement('Views');
      for i := 0 to ViewCount - 1 do
      begin
        LChildNode := LElement.CreateElement('View', LParentNode);

        View[i].SaveToXMLNode(LChildNode);
      end;
    finally
      FreeAndNil(LElement);
    end;

    LXMLDoc.SaveToStream(AStream);
  finally
    FreeAndNil(LXMLDoc);
  end;
end;

procedure TibeDataObjectDefinition.FilterClear();
begin
  FilterList.Clear();
end;

function TibeDataObjectDefinition.FilterBuild(const AFieldName: string; AFilterValue: Variant): string;
begin
  Result := FilterBuild(AFieldName, AFilterValue, '=');
end;

function TibeDataObjectDefinition.FilterBuild(const AFieldName: string; AFilterValue: Variant; AOpeator: string): string;
var
  LTableField: TibeDODefTableField;
begin
  LTableField := FFieldList.GetByFieldAlias(AFieldName);

  if not Assigned(LTableField) then
  begin
    raise EibeDataObjectDefinition.Create('TibeDataObjectDefinition.FilterBuild(): Feld "' + AFieldName + '" wurde in "' + DOClassName + '" nicht gefunden');
  end;
  
  if VarHasValue(AFilterValue) then
  begin
    Result :=  LTableField.QualifiedTblFieldName + ' ' + AOpeator + ' ' + AFilterValue;
  end else
  begin
    Result := LTableField.QualifiedTblFieldName + ' is null';
  end;
end;

procedure TibeDataObjectDefinition.FilterAdd(const AFilter: string);
begin
  FilterList.Add(AFilter);
end;

procedure TibeDataObjectDefinition.PKFilterAdd(const APKValue: string; AOmitTblAlias: Boolean);
var
  LPKField: TibeDODefTableField;
  LPKFieldName: string;
begin
  LPKField := FieldByName[MainTablePKName];
  if AOmitTblAlias then
  begin
    LPKFieldName := LPKField.Name;
  end else
  begin
    LPKFieldName := LPKField.QualifiedTblFieldName;
  end;
  
  FilterList.Add(LPKFieldName + ' = ' + APKValue);
end;

function TibeDataObjectDefinition.GetTableCount: Integer;
begin
  Result := FTableList.Count;
end;

function TibeDataObjectDefinition.GetTable(AIndex: Integer): TibeDODefTable;
begin
  Result := FTableList[AIndex];
end;

function TibeDataObjectDefinition.AddTable(): Integer;
var
  LTable: TibeDODefTable;
begin
  LTable := TibeDODefTable.Create();
  Result := FTableList.Add(LTable);
end;

function TibeDataObjectDefinition.GetFieldCount: Integer;
begin
  Result := FFieldList.Count;
end;

function TibeDataObjectDefinition.GetField(AIndex: Integer): TibeDODefTableField;
begin
  Result := FFieldList[AIndex];
end;

function TibeDataObjectDefinition.GetFieldByName(const AFieldName: string): TibeDODefTableField;
begin
  if AFieldName = '' then
  begin
    raise EibeDataObjectDefinition.Create('DataObject "' + DOClassName + '": auf Feld kann nicht mit leerem Feldnamen zugegriffen werden');
  end else
  begin
    Result := FFieldList.GetByFieldAlias(AFieldName);
  end;
end;

function TibeDataObjectDefinition.AddField(): Integer;
var
  LField: TibeDODefTableField;
begin
  LField := TibeDODefTableField.Create();
  Result := FFieldList.Add(LField);
end;

function TibeDataObjectDefinition.GetViewCount: Integer;
begin
  Result := FViewList.Count;
end;

function TibeDataObjectDefinition.GetView(AIndex: Integer): TibeDODefView;
begin
  Result := FViewList[AIndex];
end;

function TibeDataObjectDefinition.GetViewByName(const AName: string): TibeDODefView;
begin
  Result := FViewList.GetByName(AName);
end;

function TibeDataObjectDefinition.GetPostParamCount(): Integer;
begin
  Result := FPostParamList.Count;
end;

procedure TibeDataObjectDefinition.SelectActiveView(const AViewName: string);
var
  LView: TibeDODefView;
begin
  LView := FViewList.GetByName(AViewName);

  if Assigned(LView) then
  begin
    FActiveView := LView;
  end else
  begin
    raise EibeDataObjectDefinition.Create(Self, 'ActiveView kann nicht gesetzt werden, die View mit Namen "' + AViewName + '" existiert im Datenobjekt "' + DOClassName + '" nicht');
  end;
end;

function TibeDataObjectDefinition.AddView(const AName: string; AActivate: Boolean): Integer;
var
  LView: TibeDODefView;
begin
  LView := TibeDODefView.Create(AName);
  Result := FViewList.Add(LView);

  if AActivate then
  begin
    FActiveView := LView;
  end;
end;

procedure TibeDataObjectDefinition.AddViewField(const AFieldName: string; const ACaption: string; AAggFunc: TibeDODefAggFunc);
begin
  ActiveView.AddField(AFieldName, FFieldList, ACaption, AAggFunc);
end;

procedure TibeDataObjectDefinition.AddViewField(const AViewField: TibeDODefViewField);
begin
  ActiveView.AddField(AViewField, FFieldList);
end;

procedure TibeDataObjectDefinition.AddViewOrderByField(const AFieldName: string; ADirection: TibeOrderByDirection);
begin
  ActiveView.AddOrderByField(AFieldName, ADirection);
end;

procedure TibeDataObjectDefinition.AddViewGroupByField(const AFieldName: string);
begin
  if not Assigned(ActiveView.FieldByName[AFieldName]) then
  begin
    AddViewField(AFieldName);
  end;
  ActiveView.AddGroupByField(AFieldName);
end;

procedure TibeDataObjectDefinition.AddPostParam(const AFieldName: string; AValue: Variant);
var
  LField: TibeDODefTableField;
  LParam: TibeDODefPostParameter;
begin
  if AFieldName = '' then
  begin
    raise EibeDataObjectDefinition.Create('DataObject "' + DOClassName + '": Parameter kann nicht mit leerem Feldnamen angelegt werden');
  end else
  begin
    LField := FieldByName[AFieldName];

    if not Assigned(LField) then
    begin
      raise EibeDataObjectDefinition.Create('DataObject "' + DOClassName + '": Parameter fr Feld "' + AFieldName + '" kann nicht angelegt werden, das Tabellenfeld ist nicht bekannt');
    end else
    begin
      if not SameText(MainTableAlias, LField.FTblAlias) then
      begin
        raise EibeDataObjectDefinition.Create('DataObject "' + DOClassName + '": Parameter fr Feld "' + AFieldName + '" aus Tabelle "' + LField.TblAlias + '" kann nicht angelegt werden, das Tabellenfeld muss aus der Haupttabelle "' + MainTableAlias + '" kommen');
      end else
      begin
        LParam := TibeDODefPostParameter.Create(LField, AValue);
        FPostParamList.Add(LParam);
      end;
    end;
  end;
end;

procedure TibeDataObjectDefinition.ClearPostParams();
begin
  FPostParamList.Clear();
end;

procedure TibeDataObjectDefinition.SetDefinitionFilePath(const APath: TFileName);
begin
  FDefinitionFilePath := AddBackslash(APath);
end;

// ===================== [ TibeDODefTable ] ====================================

constructor TibeDODefTable.Create();
begin
  inherited Create();

  FXMElement := TibeXMLElement.Create();
end;

destructor TibeDODefTable.Destroy();
begin
  inherited Destroy();

  FreeAndNil(FXMElement);
end;

procedure TibeDODefTable.LoadFromXMLNode(ANode: IXMLNode);
begin
  Name := FXMElement.GetAttributeTypedValue('Name', '', ANode);
  Alias := FXMElement.GetAttributeTypedValue('Alias', '', ANode);
  FieldPrefix := FXMElement.GetAttributeTypedValue('FieldPrefix', '', ANode);
  JoinType := FXMElement.GetAttributeTypedValue('JoinType', '', ANode);
  JoinCondition := FXMElement.GetAttributeTypedValue('JoinCondition', '', ANode);
  Description := FXMElement.GetAttributeTypedValue('Description', '', ANode);
end;

procedure TibeDODefTable.SaveToXMLNode(ANode: IXMLNode);
begin
  FXMElement.CreateAttribute('Name', Name, ANode);
  FXMElement.CreateAttribute('Alias', Alias, ANode);
  FXMElement.CreateAttribute('FieldPrefix', FieldPrefix, ANode);
  FXMElement.CreateAttribute('JoinType', JoinType, ANode);
  FXMElement.CreateAttribute('JoinCondition', JoinCondition, ANode);
  FXMElement.CreateAttribute('Description', Description, ANode);
end;

procedure TibeDODefTable.Assign(ADefTable: TibeDODefTable);
begin
  Name := ADefTable.Name;
  Alias := ADefTable.Alias;
  FieldPrefix := ADefTable.FieldPrefix;
  JoinType := ADefTable.JoinType;
  JoinCondition := ADefTable.JoinCondition;
  Description := ADefTable.Description;
end;

function TibeDODefTableList.AddTable(): TibeDODefTable;
begin
  Result := TibeDODefTable.Create();
  Add(Result);
end;

function TibeDODefTableList.GetByAlias(const ATblAlias: string): TibeDODefTable;
var
  i: Integer;
  LTable: TibeDODefTable;
begin
  LTable := nil;

  for i := 0 to Count - 1 do
  begin
    if SameText(Self[i].Alias, ATblAlias) then
    begin
      LTable := Self[i];
      Break;
    end;
  end;

  Result := LTable;
end;

procedure TibeDODefTableList.LoadFromXMLNode(ANode: IXMLNode);
var
  i: Integer;
  LTable: TibeDODefTable;
begin
  Clear();

  for i := 0 to ANode.ChildNodes.Count - 1 do
  begin
    if ANode.ChildNodes[i].NodeName = 'Table' then
    begin
      LTable := AddTable();
      LTable.LoadFromXMLNode(ANode.ChildNodes[i]);
    end;
  end;
end;

// ======================= [ TibeDODefFieldCommon ] ============================

constructor TibeDODefFieldCommon.Create(); 
begin
  inherited Create();

  FXMElement := TibeXMLElement.Create();
end;

destructor TibeDODefFieldCommon.Destroy();
begin
  inherited Destroy();

  FreeAndNil(FXMElement);
end;

constructor TibeDODefFieldListCommon<T>.Create();
begin
  inherited Create();

  FieldNodeName := 'Field';
end;

function TibeDODefFieldListCommon<T>.AddField(): T;
begin
  Result := T.Create();
  Add(Result);
end;

procedure TibeDODefFieldListCommon<T>.LoadFromXMLNode(ANode: IXMLNode; AFieldList: TibeDODefViewFieldList);
var
  i: Integer;
  LField: T;
begin
  for i := 0 to ANode.ChildNodes.Count - 1 do
  begin
    if SameText(ANode.ChildNodes[i].NodeName, FieldNodeName) then
    begin
      LField := AddField();
      LField.LoadFromXMLNode(ANode.ChildNodes[i], AFieldList);
    end;
  end;
end;

// ===================== [ TibeDODefTableField ] ===============================

constructor TibeDODefTableField.Create();
begin
  inherited Create();

  FXMElement := TibeXMLElement.Create();
end;

destructor TibeDODefTableField.Destroy();
begin
  inherited Destroy();

  FreeAndNil(FXMElement);
end;

function TibeDODefTableField.ConvertToTableFieldValue(AValue: Variant): Variant;
var
  LValue: Variant;
begin
  if VarHasValue(AValue) then
  begin
    if DataType = dodtIntDate then
    begin
      // Sonderbehandlung fr Integer-Datum: beim zurckschreiben vom Wert wird der
      // DataTime-Wert in die Integer Repsentation umgewandelt
      LValue := DateToIntDate(AValue);
    end else if DataType = dodtIntTime then
    begin
      // Sonderbehandlung fr Integer-Zeit: beim zurckschreiben vom Wert wird der
      // DataTime-Wert in die Integer Repsentation umgewandelt
      LValue := DateToIntTime(AValue);
    end else
    begin
      // Standard: den Wert durchreichen
      LValue := AValue;
    end;
  end else
  begin
    LValue := Null;
  end;

  Result := LValue;
end;

function TibeDODefTableField.ConvertFromTableFieldValue(AValue: Variant): Variant;
begin
  if VarHasValue(AValue) then
  begin
    if (DataType = dodtIntDate) then
    begin
      // Sonderbehandlung fr Integer-Datum: beim auslesen vom Wert wird der interne
      // Integerwert in einen DataTime-Wert umgewandelt
      Result := IntDateToDate(VarAsInteger(AValue));
    end else if (DataType = dodtIntTime) then
    begin
      // Sonderbehandlung fr Integer-Zeit: beim auslesen vom Wert wird der interne
      // Integerwert in einen DataTime-Wert umgewandelt
      Result := IntTimeToDate(VarAsInteger(AValue));
    end else
    begin
      // Standard: den Wert vom Datenbankfeld durchreichen
      Result := AValue;
    end
  end else
  begin
    Result := Null;
  end;
end;

procedure TibeDODefTableField.LoadFromXMLNode(ANode: IXMLNode; ATableList: TibeDODefTableList);
begin
  TblAlias := XMElement.GetAttributeTypedValue('TblAlias', '', ANode);
  Name := FXMElement.GetAttributeTypedValue('Name', '', ANode);

  FTable := ATableList.GetByAlias(TblAlias);
  if not Assigned(Table) then
  begin
    raise EibeDataObjectDefinition.Create('Load TableField von XML: ' + TblAlias + '.' + Name + ' kann nicht erstellt werden, die Tabelle mit Alias ' + TblAlias + ' ist nicht vorhanden');
  end;

  FAlias := FXMElement.GetAttributeTypedValue('FieldAlias', '', ANode);
  if FAlias = '' then
  begin
    if IsExpression then
    begin
      FAlias := Table.FieldPrefix + '_Expression';
    end else
    begin
      FAlias := Table.FieldPrefix + '_' + Name;
    end;
  end;
  Caption := FXMElement.GetAttributeTypedValue('Caption', '', ANode);
  Expression := FXMElement.GetAttributeTypedValue('Expression', '', ANode);
  DataType := DODefDataTypeFromString(FXMElement.GetAttributeTypedValue('DataType', '', ANode));
  FWidth := FXMElement.GetAttributeTypedValue('Width', IfThen(DataType = dodtString, 255, 0), ANode);
  FDisplayOption := FXMElement.GetAttributeTypedValue('DisplayOption', '', ANode);
end;

procedure TibeDODefTableField.SaveToXMLNode(ANode: IXMLNode);
begin
  FXMElement.CreateAttribute('FieldAlias', Alias, ANode);
  FXMElement.CreateAttribute('TblAlias', TblAlias, ANode);
  FXMElement.CreateAttribute('Name', Name, ANode);
  FXMElement.CreateAttribute('Caption', Caption, ANode);
  FXMElement.CreateAttribute('Expression', Expression, ANode);
  FXMElement.CreateAttribute('DataType', DODefDataTypeIDArray[DataType], ANode);
  FXMElement.CreateAttribute('Width', Width, ANode);
  FXMElement.CreateAttribute('DisplayOption', DisplayOption, ANode);
end;

procedure TibeDODefTableField.Assign(ADefTableField: TibeDODefTableField);
begin
  TblAlias := ADefTableField.TblAlias;
  Name := ADefTableField.Name;
  Caption := ADefTableField.Caption;
  Expression := ADefTableField.Expression;
  DataType := ADefTableField.DataType;
  FWidth := ADefTableField.Width;
  FDisplayOption := ADefTableField.DisplayOption;
  FAlias := ADefTableField.Alias;
end;

function TibeDODefTableField.GetCaption(): string;
begin
  if FCaption <> '' then
  begin
    Result := FCaption;
  end else
  begin
    Result := Alias;
  end;
end;

function TibeDODefTableField.GetDBFieldClass(): TFieldClass;
begin
  case DataType of
    dodtString:   Result := TStringField;
    dodtMemo:     Result := TMemoField;
    dodtInteger,
    dodtIntDate,
    dodtIntTime:  Result := TIntegerField;
    dodtDate:     Result := TDateField;
    dodtTime:     Result := TTimeField;
    dodtDateTime: Result := TDateTimeField;
    dodtBoolean:  Result := TBooleanField;
    dodFloat:     Result := TFloatField;
    dodtMoney:    Result := TCurrencyField;
  else raise EibeDataObjectDefinition.Create('TibeDODefField.DBFieldClass fr Feld ' + Name + ' kann nicht ermittelt werden, der Datentyp wird nicht untersttzt');
  end;
end;

function TibeDODefTableField.GetQualifiedTblFieldName(): string;
begin
  if Name <> '' then
  begin
    if TblAlias <> '' then
    begin
      Result := TblAlias + '.' + Name;
    end else
    begin
      Result := Name;
    end;
  end else
  begin
    Result := '';
  end;
end;

function TibeDODefTableField.GetIsExpression(): Boolean;
begin
  Result := Trim(Expression) <> '';
end;

function TibeDODefTableFieldList.AddField(): TibeDODefTableField;
begin
  Result := TibeDODefTableField.Create();
  Add(Result);
end;

function TibeDODefTableFieldList.GetByFieldAlias(const AFieldAlias: string): TibeDODefTableField;
var
  LField: TibeDODefTableField;
begin
  Result := nil;

  for LField in Self do
  begin
    if SameText(LField.Alias, AFieldAlias) then
    begin
      Result := LField;
      Break;
    end;
  end;
end;

procedure TibeDODefTableFieldList.LoadFromXMLNode(ANode: IXMLNode; ATableList: TibeDODefTableList);
var
  i: Integer;
  LField: TibeDODefTableField;
begin
  Clear();

  for i := 0 to ANode.ChildNodes.Count - 1 do
  begin
    if ANode.ChildNodes[i].NodeName = 'Field' then
    begin
      LField := AddField();
      LField.LoadFromXMLNode(ANode.ChildNodes[i], ATableList);
    end;
  end;
end;

// ========================= [ TibeDODefView ] =================================

constructor TibeDODefView.Create();
begin
  inherited Create();
  FFieldList := TibeDODefViewFieldList.Create();
  FGroupByFieldList := TibeDODefViewGroupByFieldList.Create();
  FOrderByFieldList := TibeDODefViewOrderByFieldList.Create();
  FXMElement := TibeXMLElement.Create();

  FFieldList.OnNotify := OnFieldListChanged;
end;

constructor TibeDODefView.Create(const AName: string); 
begin
  Create();

  Name := AName;
  TouchFieldModifyTag();
end;

destructor TibeDODefView.Destroy(); 
begin
  inherited Destroy();

  FreeAndNil(FFieldList);
  FreeAndNil(FOrderByFieldList);
  FreeAndNil(FGroupByFieldList);
  FreeAndNil(FXMElement);
end;

procedure TibeDODefView.TouchFieldModifyTag();
begin
  Randomize;
  FFieldModifyTag := Random(MaxInt);
end;

procedure TibeDODefView.OnFieldListChanged(Sender: TObject; const Item: TibeDODefViewField; Action: TCollectionNotification);
begin
  // mit jeder nderung der Feldliste einen neuen Wert in ModifyTag schreiben.
  // Damit kann ein Nutzer der View erkennen, dass die Feldliste seit seiner 
  // letzten Verwendung selbiger gendert wurde
  TouchFieldModifyTag();
end;

procedure TibeDODefView.LoadFromXMLNode(ANode: IXMLNode; AFieldList: TibeDODefTableFieldList);
var
  i: Integer;
begin
  Name := FXMElement.GetAttributeTypedValue('Name', '', ANode);
  Description := FXMElement.GetAttributeTypedValue('Description', '', ANode);
  ViewPKName := FXMElement.GetAttributeTypedValue('ViewPKName', '', ANode);

  for i := 0 to ANode.ChildNodes.Count - 1 do
  begin
    if ANode.ChildNodes[i].NodeName = 'Fields' then
    begin
      FFieldList.LoadFromXMLNode(ANode.ChildNodes[i], AFieldList);
    end;

    if ANode.ChildNodes[i].NodeName = 'GroupBy' then
    begin
      FGroupByFieldList.LoadFromXMLNode(ANode.ChildNodes[i], FFieldList);
    end;

    if ANode.ChildNodes[i].NodeName = 'OrderBy' then
    begin
      FOrderByFieldList.LoadFromXMLNode(ANode.ChildNodes[i], FFieldList);
    end;
  end;
end;

procedure TibeDODefView.SaveToXMLNode(ANode: IXMLNode);
var
  LChildNode: IXMLNode;
begin
  FXMElement.CreateAttribute('Name', Name, ANode);
  FXMElement.CreateAttribute('Description', Description, ANode);
  FXMElement.CreateAttribute('ViewPKName', ViewPKName, ANode);

  if FieldCount > 0 then
  begin
    LChildNode := FXMElement.CreateElement('Fields', ANode);
    FFieldList.SaveToXMLNode(LChildNode);
  end;

  if OrderByFieldCount > 0 then
  begin
    LChildNode := FXMElement.CreateElement('OrderBy', ANode);
    FOrderByFieldList.SaveToXMLNode(LChildNode);
  end;

  if GroupByFieldCount > 0 then
  begin
    LChildNode := FXMElement.CreateElement('GroupBy', ANode);
    FGroupByFieldList.SaveToXMLNode(LChildNode);
  end;
end;

procedure TibeDODefView.CreateDatasetStructure(ADat: TDataSet);
var
  LViewField: TibeDODefViewField;
begin
  ADat.Fields.Clear();
  ADat.FieldDefs.Clear();
  //ADat.FieldDefs.Update();

  for LViewField in FFieldList do
  begin
    LViewField.AddFieldToDataset(ADat);
  end;
end;

procedure TibeDODefView.AddField(const ATblFieldAlias: string; ATblFieldList: TibeDODefTableFieldList; const ACaption: string; AAggFunc: TibeDODefAggFunc);
var
  LViewField: TibeDODefViewField;
begin
  LViewField := FFieldList.AddField();
  LViewField.InitFieldAttributes(ATblFieldAlias, AAggFunc, ATblFieldList);

  if ACaption <> '' then
  begin
    LViewField.Caption := ACaption;
  end;
end;

procedure TibeDODefView.AddField(AViewField: TibeDODefViewField; ATblFieldList: TibeDODefTableFieldList);
var
  LViewField: TibeDODefViewField;
begin
  LViewField := FFieldList.AddField();
  //LViewField.InitFieldAttributes(AViewField.TblFieldAlias, AViewField.AggFunc, ATblFieldList);
  LViewField.Assign(AViewField, ATblFieldList);
end;

//procedure TibeDODefView.AddCountAllField(const ATblAlias, AFieldName: string; const ACaption: string = '');
//var
//  LViewField: TibeDODefViewField;
//begin
//  LViewField := TibeDODefViewField.CreateCountAll(ATblAlias, AFieldName, ACaption);
//  FFieldList.Add(LViewField);
//end;

procedure TibeDODefView.AddOrderByField(const ATblFieldAlias: string; ADirection: TibeOrderByDirection);
var
  LField: TibeDODefViewOrderByField;
begin
  LField := FOrderByFieldList.AddField();
  LField.InitFieldAttributes(ATblFieldAlias, FFieldList, ADirection);
end;

procedure TibeDODefView.AddGroupByField(const ATblFieldAlias: string);
var
  LField: TibeDODefViewGroupByField;
begin
  LField := FGroupByFieldList.AddField();
  LField.InitFieldAttributes(ATblFieldAlias, FFieldList);
end;

procedure TibeDODefView.SetName(const AName: string);
begin
  FName := AName;
end;

function TibeDODefView.GetFieldCount(): Integer;
begin
  Result := FFieldList.Count;
end;

function TibeDODefView.GetField(AIndex: Integer): TibeDODefViewField;
begin
  Result := FFieldList[AIndex];
end;

function TibeDODefView.GetFieldByName(const AFieldName: string): TibeDODefViewField;
var
  LField: TibeDODefViewField;
begin
  Result := nil;

  for LField in FFieldList do
  begin
    if SameText(AFieldName, LField.FieldName) then
    begin
      Result := LField;
      Break;
    end;
  end;
end;

function TibeDODefView.GetOrderByFieldCount(): Integer;
begin
  Result := FOrderByFieldList.Count;
end;

function TibeDODefView.GetOrderByField(AIndex: Integer): TibeDODefViewOrderByField;
begin
  Result := FOrderByFieldList[AIndex];
end;

function TibeDODefView.GetGroupByFieldCount(): Integer;
begin
  Result := FGroupByFieldList.Count;
end;

function TibeDODefView.GetGroupByField(AIndex: Integer): TibeDODefViewGroupByField;
begin
  Result := FGroupByFieldList[AIndex];
end;

function TibeDODefViewList.GetByName(const AViewName: string): TibeDODefView;
var
  i: Integer;
  LView: TibeDODefView;
begin
  LView := nil;

  for i := 0 to Count - 1 do
  begin
    if SameText(Self[i].Name, AViewName) then
    begin
      LView := Self[i];
    end;
  end;

  Result := LView;
end;

function TibeDODefViewList.AddView(): TibeDODefView;
begin
  Result := TibeDODefView.Create();
  Add(Result);
end;

procedure TibeDODefViewList.LoadFromXMLNode(ANode: IXMLNode; AFieldList: TibeDODefTableFieldList);
var
  i: Integer;
  LView: TibeDODefView;
begin
  Clear();

  for i := 0 to ANode.ChildNodes.Count - 1 do
  begin
    if ANode.ChildNodes[i].NodeName = 'View' then
    begin
      LView := AddView();
      LView.LoadFromXMLNode(ANode.ChildNodes[i], AFieldList);
    end;
  end;
end;

// ===================== [ TibeDODefViewField ] ===============================

constructor TibeDODefViewField.Create();
begin
  inherited Create();

  FXMElement := TibeXMLElement.Create();
end;

constructor TibeDODefViewField.Create(ATblField: TibeDODefTableField; const ACaption: string = ''; AAggFunc: TibeDODefAggFunc = doafNone); 
begin
  Create();

  FTblFieldAlias := ATblField.Alias;
  VisibleInList := True;
  FCaption := ACaption;
  FAggFunc := AAggFunc;

  InitFieldAttributesFromTblField(ATblField);
end;

constructor TibeDODefViewField.CreateCountAll(const ATblAlias, AFieldName: string; ACaption: string = '');
begin
  Create();

  FTblFieldAlias := ATblAlias;
  VisibleInList := True;
  FCaption := ACaption;
  FAggFunc := doafCount;
  ColumnWidth := 100;
  FTblField := nil;
end;

destructor TibeDODefViewField.Destroy();
begin
  inherited Destroy();

  FreeAndNil(FXMElement);
end;

procedure TibeDODefViewField.InitFieldAttributes(const ATblFieldAlias: string; AAggFunc: TibeDODefAggFunc; AFieldList: TibeDODefTableFieldList);
begin
  if ATblFieldAlias = '' then
  begin
    raise EibeDataObjectDefinition.Create('Init ViewField Attributes: Feld kann nicht erstellt werden, es ist kein Alias fr das Tabellenfeld angegeben');
  end;

  FTblField := AFieldList.GetByFieldAlias(ATblFieldAlias);
  if Assigned(TblField) then
  begin
    InitFieldAttributesFromTblField(TblField);
  end else
  begin
    // wenn das Feld nicht als Count-Aggregat angelegt wird muss zwingend eine
    // Referenz auf das Tabellenfeld existieren. Im Fehlerfall hie eine Exception
    // auslsen
    if AAggFunc <> doafCount then
    begin
      raise EibeDataObjectDefinition.Create('Init ViewField Attributes: Feld kann nicht erstellt werden, fr den Tabellenfeld-Alias "' + ATblFieldAlias + '" ist kein Tabellenfeld vorhanden');
    end;
  end;

  FTblFieldAlias := ATblFieldAlias;
  FieldName := TblFieldAlias;
  FAggFunc := AAggFunc;
  case AggFunc of
    doafCount: FieldName := FieldName + '_Count';
    doafSum:   FieldName := FieldName + '_Sum';
    doafAvg:   FieldName := FieldName + '_Avg';
    doafMin:   FieldName := FieldName + '_Min';
    doafMax:   FieldName := FieldName + '_Max';
  end;
end;

procedure TibeDODefViewField.InitFieldAttributesFromTblField(ATblField: TibeDODefTableField);
begin
  FTblField := ATblField;

  if ColumnWidth = 0 then
  begin
    // im Default die Spaltenbreite aus dem Tabellenfeld berechnen
    case ATblField.DataType of
      dodtString:   ColumnWidth := Min(3 * ATblField.Width, 250);
      dodtInteger:  ColumnWidth := 75;
      dodtIntDate:  ColumnWidth := 75;
      dodtIntTime:  ColumnWidth := 75;
      dodtDate:     ColumnWidth := 75;
      dodtTime:     ColumnWidth := 50;
      dodtDateTime: ColumnWidth := 125;
      dodtBoolean:  ColumnWidth := 35;
      dodFloat:     ColumnWidth := 100;
      dodtMoney:    ColumnWidth := 100;
      dodtMemo:     ColumnWidth := 100;
    end;
  end;
end;

procedure TibeDODefViewField.LoadFromXMLNode(ANode: IXMLNode; AFieldList: TibeDODefTableFieldList);
var
  LTblFieldAlias, LAggFuncID, LDataType: string;
begin
  LTblFieldAlias := Trim(FXMElement.GetAttributeTypedValue('TblFieldAlias', '', ANode));
  LAggFuncID := Trim(FXMElement.GetAttributeTypedValue('Aggregat', '', ANode));
  ColumnWidth := FXMElement.GetAttributeTypedValue('ColumnWidth', 0, ANode);

  InitFieldAttributes(LTblFieldAlias, DODefAggFuncTypeFromString(LAggFuncID), AFieldList);

  LDataType := FXMElement.GetAttributeTypedValue('DataType', DODefDataTypeIDArray[TblDataType], ANode);
  FDataType := DODefDataTypeFromString(LDataType);
  VisibleInList := FXMElement.GetAttributeTypedValue('VisibleInList', True, ANode);
  FCaption := Trim(FXMElement.GetAttributeTypedValue('Caption', Caption, ANode));
end;

procedure TibeDODefViewField.SaveToXMLNode(ANode: IXMLNode);
begin
  FXMElement.CreateAttribute('TblFieldAlias', TblFieldAlias, ANode);
  FXMElement.CreateAttribute('Aggregat', DODefAggFuncIDArray[AggFunc], ANode);
  FXMElement.CreateAttribute('DataType', DODefDataTypeIDArray[TblDataType], ANode);
  FXMElement.CreateAttribute('ColumnWidth', ColumnWidth, ANode);
  FXMElement.CreateAttribute('VisibleInList', VisibleInList, ANode);
  FXMElement.CreateAttribute('Caption', Caption, ANode);
end;

procedure TibeDODefViewField.Assign(ADefViewField: TibeDODefViewField; ATblFieldList: TibeDODefTableFieldList);
begin
  FTblFieldAlias := ADefViewField.TblFieldAlias;
  FAggFunc := ADefViewField.AggFunc;
  ColumnWidth := ADefViewField.ColumnWidth;

  InitFieldAttributes(TblFieldAlias, FAggFunc, ATblFieldList);

  //FTblField := ADefViewField.TblField;
  FieldName := ADefViewField.FieldName;
  Caption := ADefViewField.Caption;
  FDataType := ADefViewField.DataType;
  VisibleInList := ADefViewField.VisibleInList;
end;

function TibeDODefViewField.AddFieldToDataset(ADataset: TDataSet): TField;
begin
  Result := TblFieldClass.Create(ADataSet);
  Result.FieldName := FieldName;
  Result.DisplayLabel := Caption;
  Result.Size := IfThen(Result is TStringField, Width, 0);
  Result.Visible := VisibleInList;
  Result.DataSet := ADataSet;
  if Result is TNumericField then
  begin
    TNumericField(Result).DisplayFormat := DisplayOption;
  end else if Result is TDateTimeField then
  begin
    TDateTimeField(Result).DisplayFormat := DisplayOption;
  end else if Result is TBooleanField then
  begin
    TBooleanField(Result).DisplayValues := DisplayOption;
  end;
end;

function TibeDODefViewField.GetSQLExpressionText: string;
var
  LColumnExpression: string;
begin
  if Assigned(TblField) then
  begin
    if TblField.IsExpression then
    begin
      LColumnExpression := TblField.Expression;
    end else
    begin
      LColumnExpression := TblField.QualifiedTblFieldName;
    end;
  end else
  begin
    LColumnExpression := '';
  end;

  case AggFunc of
    doafCount: Result := 'count(*)';
    doafSum:   Result := 'sum(' + LColumnExpression + ')'; 
    doafAvg:   Result := 'avg(' + LColumnExpression + ')'; 
    doafMin:   Result := 'min(' + LColumnExpression + ')'; 
    doafMax:   Result := 'max(' + LColumnExpression + ')';
  else         Result := LColumnExpression;
  end;

  if Result = '' then
  begin
    raise EibeDataObjectDefinition.Create('Fehler beim Erstellen von SQL-Ausdruck fr ViewField. Von den Angaben Tabellenfeldname, Feldausdruck oder Count-Aggregat uss mindestens eine bekannt sein, es ist aber keine Angabe gefllt');
  end;
end;
  
function TibeDODefViewField.GetSQLColumnText: string;
begin
  Result := SQLExpressionText + ' as ' + FieldName;
end;

procedure TibeDODefViewField.SetVisibleInList(AVisible: Boolean);
begin
  FVisibleInList := AVisible;
end;

function TibeDODefViewField.GetTblFieldClass(): TFieldClass;
begin
  if AggFunc <> doafCount then
  begin
    //Result := TblField.DBFieldClass;
    Result := FieldClass;
  end else
  begin 
    Result := TIntegerField;
  end;
end;

function TibeDODefViewField.GetTblDataType(): TibeDODefDataType;
begin
  Result := TblField.DataType;
end;

function TibeDODefViewField.GetTblFieldCaption(): string;
begin
  if Assigned(TblField) then
  begin
    Result := TblField.Caption;
  end else
  begin
    Result := FieldName;
  end;
end;

function TibeDODefViewField.GetTblAlias(): string;
begin
  Result := TblField.TblAlias;
end;

function TibeDODefViewField.GetFieldClass(): TFieldClass;
begin
  case DataType of
    dodtString:   Result := TStringField;
    dodtMemo:     Result := TMemoField;
    dodtInteger:  Result := TIntegerField;
    dodtIntDate,
    dodtDate:     Result := TDateField;
    dodtIntTime,
    dodtTime:     Result := TTimeField;
    dodtDateTime: Result := TDateTimeField;
    dodtBoolean:  Result := TBooleanField;
    dodFloat:     Result := TFloatField;
    dodtMoney:    Result := TCurrencyField;
  else raise EibeDataObjectDefinition.Create('TibeDODefViewField.DBFieldClass fr Feld ' + FieldName + ' kann nicht ermittelt werden, der Datentyp wird nicht untersttzt');
  end;
end;

function TibeDODefViewField.GetDataType(): TibeDODefDataType;
begin
  if AggFunc <> doafCount then
  begin
    Result := FDataType;
  end else
  begin 
    Result := dodtInteger;
  end;
end;

function TibeDODefViewField.GetWidth(): Integer;
begin
  Result := TblField.Width;
end;

function TibeDODefViewField.GetDisplayOption(): string;
begin
  Result := TblField.DisplayOption;
end;

function TibeDODefViewFieldList.AddField(): TibeDODefViewField;
begin
  Result := TibeDODefViewField.Create();
  Add(Result);
end;

procedure TibeDODefViewFieldList.LoadFromXMLNode(ANode: IXMLNode; AFieldList: TibeDODefTableFieldList);
var
  i: Integer;
  LField: TibeDODefViewField;
begin
  for i := 0 to ANode.ChildNodes.Count - 1 do
  begin
    if ANode.ChildNodes[i].NodeName = 'Field' then
    begin
      LField := AddField();
      LField.LoadFromXMLNode(ANode.ChildNodes[i], AFieldList);
    end;
  end;
end;

procedure TibeDODefViewFieldList.SaveToXMLNode(ANode: IXMLNode);
var
  LElement: TibeXMLElement;
  LChildNode: IXMLNode;
  LField: TibeDODefViewField;
begin
  LElement := TibeXMLElement.Create();
  try
    for LField in Self do
    begin
      LChildNode := LElement.CreateElement('Field', ANode);
      LField.SaveToXMLNode(LChildNode);
    end;
  finally
    FreeAndNil(LElement);
  end;
end;

function TibeDODefViewFieldList.GetFieldByName(const AFieldName: string): TibeDODefViewField;
var
  LField: TibeDODefViewField;
begin
  Result := nil;

  for LField in Self do
  begin
    if SameText(AFieldName, LField.FieldName) then
    begin
      Result := LField;
      Break;
    end;
  end;
end;

// ===================== [ TibeDODefViewOrderByField ] =========================

procedure TibeDODefViewOrderByFieldList.SaveToXMLNode(ANode: IXMLNode);
var
  LElement: TibeXMLElement;
  LChildNode: IXMLNode;
  LField: TibeDODefViewOrderByField;
begin
  LElement := TibeXMLElement.Create();
  try
    for LField in Self do
    begin
      LChildNode := LElement.CreateElement('Field', ANode);
      LField.SaveToXMLNode(LChildNode);
    end;
  finally
    FreeAndNil(LElement);
  end;
end;

procedure TibeDODefViewOrderByField.LoadFromXMLNode(ANode: IXMLNode; AFieldList: TibeDODefViewFieldList);
var
  LTblFieldAlias, LDirection: string;
  LDirectionType: TibeOrderByDirection;
begin
  LTblFieldAlias := XMElement.GetAttributeTypedValue('TblFieldAlias', '', ANode);

  LDirection := XMElement.GetAttributeTypedValue('Direction', DODefOrderByIDArray[obdAscending], ANode);
  if LDirection <> '' then
  begin
    LDirectionType := DODefOrderByTypeFromString(LDirection);
  end else
  begin
    raise EibeDataObjectDefinition.Create('DO-Definition: nicht untersttze Sortierrichtung "' + LDirection + '" fr Order By Feld "' + TblFieldAlias + '"');
  end;

  InitFieldAttributes(LTblFieldAlias, AFieldList, LDirectionType);
end;

procedure TibeDODefViewOrderByField.SaveToXMLNode(ANode: IXMLNode);
begin
  XMElement.CreateAttribute('TblFieldAlias', TblFieldAlias, ANode);
  XMElement.CreateAttribute('Direction', DODefOrderByIDArray[Direction], ANode);
end;

procedure TibeDODefViewOrderByField.Assign(AOrderByField: TibeDODefViewOrderByField);
begin
  FTblFieldAlias := AOrderByField.TblFieldAlias;
  FDirection := AOrderByField.Direction;
end;

procedure TibeDODefViewOrderByField.InitFieldAttributes(const ATblFieldAlias: string; AFieldList: TibeDODefViewFieldList; ADirection: TibeOrderByDirection);
begin
  FTblFieldAlias := ATblFieldAlias;
  FDirection := ADirection;
  FViewField := AFieldList.FieldByName[TblFieldAlias];
end;

function TibeDODefViewOrderByField.GetSQLText(): string;
begin
  if Assigned(FViewField) then
  begin
    Result := FViewField.SQLExpressionText;
  end else
  begin
    Result := TblFieldAlias;
  end;

  case Direction of
    obdAscending: Result := Result + ' asc';
    obdDescending: Result := Result + ' desc';
  end;
end;

// ===================== [ TibeDODefViewGroupByField ] =========================

procedure TibeDODefViewGroupByFieldList.SaveToXMLNode(ANode: IXMLNode);
var
  LElement: TibeXMLElement;
  LChildNode: IXMLNode;
  LField: TibeDODefViewGroupByField;
begin
  LElement := TibeXMLElement.Create();
  try
    for LField in Self do
    begin
      LChildNode := LElement.CreateElement('Field', ANode);
      LField.SaveToXMLNode(LChildNode);
    end;
  finally
    FreeAndNil(LElement);
  end;
end;

procedure TibeDODefViewGroupByField.LoadFromXMLNode(ANode: IXMLNode; AFieldList: TibeDODefViewFieldList);
var
  LTblFieldAlias: string;
begin
  LTblFieldAlias := XMElement.GetAttributeTypedValue('TblFieldAlias', '', ANode);
  InitFieldAttributes(LTblFieldAlias, AFieldList);
end;

procedure TibeDODefViewGroupByField.SaveToXMLNode(ANode: IXMLNode);
begin
  FXMElement.CreateAttribute('TblFieldAlias', TblFieldAlias, ANode);
end;

procedure TibeDODefViewGroupByField.Assign(AGroupByField: TibeDODefViewGroupByField);
begin
  FTblFieldAlias := AGroupByField.TblFieldAlias;
end;

procedure TibeDODefViewGroupByField.InitFieldAttributes(const ATblFieldAlias: string; AFieldList: TibeDODefViewFieldList);
begin
  FTblFieldAlias := ATblFieldAlias;
  FViewField := AFieldList.FieldByName[TblFieldAlias];
  if not Assigned(ViewField) then
  begin
    raise EibeDataObjectDefinition.Create('DO-Definition: Group By Feld "' + TblFieldAlias + '" kann nicht hinzu gefgt werden, das Feld ist in der View nicht definiert');
  end;
end;

function TibeDODefViewGroupByField.GetSQLText(): string;
begin
  Result := ViewField.SQLExpressionText;
end;

// ====================== [ TibeDODefFilter ] ==================================

constructor TibeDODefFilter.Create(const AFilter: string);
begin
  inherited Create();

  Filter := AFilter;
end;

function TibeDODefFilterList.Add(const AFilter: string): Integer;
begin
  Result := inherited Add(TibeDODefFilter.Create(AFilter));
end;

// ================ [ TibeDODefPostParameter ] =================================

constructor TibeDODefPostParameter.Create(ATblField: TibeDODefTableField; AValue: Variant);
begin
  inherited Create();

  FTblField := ATblField;
  FValue := AValue;
end;

function TibeDODefPostParameter.GetFieldName(): string;
begin
  Result := FTblField.Alias;
end;

function TibeDODefPostParameter.GetTblFieldName(): string;
begin
  Result := FTblField.Name;
end;

end.
