// ============================================================================
//
// Unit:    ibeDataObjectFields
// Inhalt:  Formular
// Aufgabe: Klassenbibliothek fr DataObject Fields
//
//   (c) by Klaus Etscheid since 2015
// ============================================================================
unit ibeDataObjectFields;

interface

uses
  SysUtils, Variants, Generics.Defaults, Generics.Collections, DB,
  ibeVariant, ibeFuncDateUtil,
  ibeDataObjectTypes, ibeDataObjectDefinition;

type

TibeDataObjectInternalField = class;

TOnFieldValueChanged = procedure(AField: TibeDataObjectInternalField) of object;

// internes Feldobjekt mit vollem Zugriff auf alle Eigenschaften
TibeDataObjectInternalField = class
private
  FOnValueChanged: TOnFieldValueChanged;
  FViewField: TibeDODefViewField;
  FDBField: TField;
  FValueOrig: Variant;
  FModified: Boolean;
  function GetFieldName: string;
  function GetValue(): Variant;
  procedure SetValue(AValue: Variant);
  function GetHasValue(): Boolean;
  function GetDBFieldIndex(): Integer;
  function GetTblAlias(): string;
protected
  property ViewField: TibeDODefViewField read FViewField;
  property DBField: TField read FDBField;
public
  constructor Create(AViewField: TibeDODefViewField; ADBField: TField; AOnValueChanged: TOnFieldValueChanged);
  procedure InitFromDBField();
  function AddDBFieldToDataset(ADataset: TDataSet): TField;
  procedure ExchangeDBField(AField: TField);
  property TblAlias: string read GetTblAlias;
  property FieldName: string read GetFieldName;
  property Value: Variant read GetValue write SetValue;
  property HasValue: Boolean read GetHasValue;
  property ValueOrig: Variant read FValueOrig;
  property Modified: Boolean read FModified;
  property DBFieldIndex: Integer read GetDBFieldIndex;
  function ConvertToTableFieldValue(AValue: Variant): Variant;
  function ConvertFromTableFieldValue(AValue: Variant): Variant;
  property OnValueChanged: TOnFieldValueChanged read FOnValueChanged write FOnValueChanged;
end;

TibeDataObjectInternalFieldList =class(TObjectList<TibeDataObjectInternalField>)
private
  FDefFieldsModifiedTag: Integer;
  function GetFieldByName(const AFieldName: string): TibeDataObjectInternalField;
protected
public
  function AddField(AViewField: TibeDODefViewField; ADBField: TField; AOnValueChanged: TOnFieldValueChanged): TibeDataObjectInternalField;
  procedure InitFieldsFromDataset();
  procedure BindDataSetFields(ADataSet: TDataSet);
  procedure AddFieldsToDataset(ADataSet: TDataSet);
  procedure AddDBFieldsToDataset(ADataSet: TDataSet);
  procedure ExchangeDataSet(ADataSet: TDataSet);
  property FieldByName[const AFieldName: string]: TibeDataObjectInternalField read GetFieldByName;
  property DefFieldsModifiedTag: Integer read FDefFieldsModifiedTag write FDefFieldsModifiedTag;
end;

// ffentliches Datenobjekt-Feld mit eingeschrnktem Zugriff auf verffentlichte 
// Eigenschaften vom internen Feld
TibeDataObjectField = class
private
  FInternalField: TibeDataObjectInternalField;
  function GetFieldName(): string;
  function GetCaption(): string;
  function GetColumnWidth(): Integer;
  function GetVisibleInList(): Boolean;
  function GetValue(): Variant;
  procedure SetValue(AValue: Variant);
  function GetHasValue(): Boolean;
  function GetIsInternalFieldAssigned(): Boolean;
protected
public
  procedure SetInternalField(AInternalField: TibeDataObjectInternalField);
  property IsInternalFieldAssigned: Boolean read GetIsInternalFieldAssigned;
  property FieldName: string read GetFieldName;
  property Caption: string read GetCaption;
  property ColumnWidth: Integer read GetColumnWidth;
  property VisibleInList: Boolean read GetVisibleInList;
  property Value: Variant read GetValue write SetValue;
  property HasValue: Boolean read GetHasValue;
  function ConvertToTableFieldValue(AValue: Variant): Variant;
  function ConvertFromTableFieldValue(AValue: Variant): Variant;
end;

implementation

// ======================= [ TibeDataObjectInternalField ] =====================

constructor TibeDataObjectInternalField.Create(AViewField: TibeDODefViewField; ADBField: TField; AOnValueChanged: TOnFieldValueChanged);
begin
  inherited Create();

  FViewField := AViewField;
  FDBField := ADBField;
  if Assigned(ADBField) then
  begin
    DBField.DisplayLabel := AViewField.Caption;
  end;
  FOnValueChanged := AOnValueChanged;
  FModified := False;
end;

procedure TibeDataObjectInternalField.InitFromDBField();
begin
  FValueOrig := DBField.Value;
  FModified := False;
end;

function TibeDataObjectInternalField.AddDBFieldToDataset(ADataset: TDataSet): TField;
var
  LFieldClass: TFieldClass;
begin
  LFieldClass := TFieldClass(DBField.ClassType);
  if LFieldClass = TAutoIncField then
  begin
    LFieldClass := TIntegerField;
  end;

  Result := LFieldClass.Create(ADataSet);
  Result.FieldName := DBField.FieldName;
  Result.DisplayLabel := DBField.DisplayLabel;
  Result.Size := IfThen(Result is TStringField, TStringField(DBField).Size, 0);
  Result.Visible := ViewField.VisibleInList;
  Result.DataSet := ADataSet;
  if Result is TNumericField then
  begin
    TNumericField(Result).DisplayFormat := ViewField.DisplayOption;
  end else if Result is TDateTimeField then
  begin
    TDateTimeField(Result).DisplayFormat := ViewField.DisplayOption;
  end else if Result is TBooleanField then
  begin
    TBooleanField(Result).DisplayValues := ViewField.DisplayOption;
  end;
end;

procedure TibeDataObjectInternalField.ExchangeDBField(AField: TField);
begin
  FDBField := AField;
end;

function TibeDataObjectInternalField.GetFieldName: string;
begin
//  Result := DBField.FieldName;
  Result := ViewField.FieldName;
end;

function TibeDataObjectInternalField.GetDBFieldIndex(): Integer;
begin
  Result := DBField.Index;
end;

function TibeDataObjectInternalField.GetTblAlias(): string;
begin
  Result := '';

  if Assigned(ViewField) then
  begin
    Result := ViewField.TblAlias;
  end;
end;

function TibeDataObjectInternalField.GetValue(): Variant;
begin
  Result := ConvertFromTableFieldValue(DBField.Value);
end;

procedure TibeDataObjectInternalField.SetValue(AValue: Variant);
begin
  if not VarsAreEqual(AValue, Value) then
  begin
    FModified := True;

    if VarHasValue(AValue) then
    begin
      DBField.Value := ConvertToTableFieldValue(AValue);
    end else
    begin
      DBField.Clear();
    end;

    if Assigned(FOnValueChanged) then
    begin
      FOnValueChanged(Self);
    end;
  end;
end;

function TibeDataObjectInternalField.GetHasValue(): Boolean;
begin
  Result := not DBField.IsNull;
end;

function TibeDataObjectInternalField.ConvertToTableFieldValue(AValue: Variant): Variant;
var
  LValue: Variant;
begin
  if VarHasValue(AValue) then
  begin
    if (ViewField.TblDataType = dodtIntDate) then
    begin
      // Sonderbehandlung fr Integer-Datum: beim zurckschreiben vom Wert wird der
      // DataTime-Wert in die Integer Repsentation umgewandelt
      LValue := DateToIntDate(AValue);
    end else if (ViewField.TblDataType = dodtIntTime) then
    begin
      // Sonderbehandlung fr Integer-Zeit: beim zurckschreiben vom Wert wird der
      // DataTime-Wert in die Integer Repsentation umgewandelt
      LValue := DateToIntTime(AValue);
    end else
    begin
      // Standard: den Wert durchreichen
      LValue := AValue;
    end;
  end else
  begin
    LValue := Null;
  end;

  Result := LValue;
end;

function TibeDataObjectInternalField.ConvertFromTableFieldValue(AValue: Variant): Variant;
begin
  if VarHasValue(AValue) then
  begin
    if (ViewField.TblDataType = dodtIntDate) then
    begin
      // Sonderbehandlung fr Integer-Datum: beim auslesen vom Wert wird der interne
      // Integerwert in einen DataTime-Wert umgewandelt
      Result := IntDateToDate(VarAsInteger(AValue));
    end else if (ViewField.TblDataType = dodtIntTime) then
    begin
      // Sonderbehandlung fr Integer-Zeit: beim auslesen vom Wert wird der interne
      // Integerwert in einen DataTime-Wert umgewandelt
      Result := IntTimeToDate(VarAsInteger(AValue));
    end else
    begin
      // Standard: den Wert vom Datenbankfeld durchreichen
      Result := AValue;
    end
  end else
  begin
    Result := Null;
  end;
end;

// ============== [ TibeDataObjectInternalFieldList ] ==========================

function TibeDataObjectInternalFieldList.AddField(AViewField: TibeDODefViewField; ADBField: TField; AOnValueChanged: TOnFieldValueChanged): TibeDataObjectInternalField;
begin
  Result := TibeDataObjectInternalField.Create(AViewField, ADBField, AOnValueChanged);
  Add(Result);
end;

procedure TibeDataObjectInternalFieldList.InitFieldsFromDataset();
var
  LField: TibeDataObjectInternalField;
begin
  for LField in Self do
  begin
    LField.InitFromDBField();
  end;
end;

procedure TibeDataObjectInternalFieldList.BindDataSetFields(ADataSet: TDataSet);
var
  LField: TibeDataObjectInternalField;
begin
  for LField in Self do
  begin
    LField.ExchangeDBField(ADataSet.FieldByName(LField.FieldName));
  end;
end;

procedure TibeDataObjectInternalFieldList.AddFieldsToDataset(ADataSet: TDataSet);
var
  LDOField: TibeDataObjectInternalField;
begin
  ADataSet.Fields.Clear();
  ADataSet.FieldDefs.Clear();

  for LDOField in Self do
  begin
    LDOField.ViewField.AddFieldToDataset(ADataSet);
  end;
end;

procedure TibeDataObjectInternalFieldList.AddDBFieldsToDataset(ADataSet: TDataSet);
var
  LDOField: TibeDataObjectInternalField;
begin
  ADataSet.Fields.Clear();
  ADataSet.FieldDefs.Clear();

  for LDOField in Self do
  begin
    LDOField.AddDBFieldToDataset(ADataSet);
  end;
end;

procedure TibeDataObjectInternalFieldList.ExchangeDataSet(ADataSet: TDataSet);
begin
  BindDataSetFields(ADataSet);
end;

function TibeDataObjectInternalFieldList.GetFieldByName(const AFieldName: string): TibeDataObjectInternalField;
var
  i: Integer;
begin
  if AFieldName = '' then
  begin
    raise EibeDataObjectDefinition.Create('DataObject FieldList kann Feld mit leerem Feldnamen nicht suchen');
  end else
  begin
    Result := nil;
  
    for i := 0 to Count - 1 do
    begin
      if SameText(AFieldName, Self[i].FieldName) then
      begin    
        Result := Self[i];
        break;
      end;
    end;
  end;
end;

// ============================== [ TibeDataObjectField ] ======================

function TibeDataObjectField.GetFieldName: string;
begin
  Result := FInternalField.FieldName;
end;

function TibeDataObjectField.GetCaption(): string;
begin
  Result := IfThen(FInternalField.ViewField.Caption <> '', FInternalField.ViewField.Caption, FInternalField.ViewField.TblFieldCaption);
end;

function TibeDataObjectField.GetColumnWidth(): Integer;
begin
  Result := FInternalField.ViewField.ColumnWidth;
end;

function TibeDataObjectField.GetVisibleInList(): Boolean;
begin
  Result := FInternalField.ViewField.VisibleInList;
end;

function TibeDataObjectField.GetValue: Variant;
begin
  Result := FInternalField.Value;
end;

procedure TibeDataObjectField.SetValue(AValue: Variant);
begin
  FInternalField.Value := AValue;
end;

function TibeDataObjectField.GetHasValue(): Boolean;
begin
  Result := FInternalField.HasValue;
end;

function TibeDataObjectField.ConvertToTableFieldValue(AValue: Variant): Variant;
begin
  Result := FInternalField.ConvertToTableFieldValue(AValue);
end;

function TibeDataObjectField.ConvertFromTableFieldValue(AValue: Variant): Variant;
begin
  Result := FInternalField.ConvertFromTableFieldValue(AValue);
end;

procedure TibeDataObjectField.SetInternalField(AInternalField: TibeDataObjectInternalField);
begin
  FInternalField := AInternalField;
end;

function TibeDataObjectField.GetIsInternalFieldAssigned(): Boolean;
begin
  Result := Assigned(FInternalField);
end;

end.
