// ============================================================================
//
// Unit:    ibeDataObjectTypes
// Inhalt:  Formular
// Aufgabe: Typbibliothek fr DataObjects
//
//   (c) by Klaus Etscheid since 2015
// ============================================================================
unit ibeDataObjectTypes;

interface

uses
  SysUtils, Data.DB,
  ibeException;

type

TibeDataObjectState = (dosNone, dosBrowse, dosInsert, dosEdit);
TibeOrderByDirection = (obdAscending, obdDescending);

const DODefOrderByIDArray: array[TibeOrderByDirection] of string = ('ASC', 'DESC');

function DODefOrderByTypeFromString(const AOrderByID: string): TibeOrderByDirection;

type

TibeDODefDataType = (dodtString, dodtInteger, dodtIntDate, dodtIntTime, dodtDate, dodtTime,
                     dodtDateTime, dodtBoolean, dodFloat, dodtMoney, dodtMemo);

const DODefDataTypeIDArray: array[TibeDODefDataType] of string = ('String', 'Int', 'IntDate', 'IntTime', 'Date', 'Time', 'DateTime', 'Boolean', 'Float', 'Money', 'Memo');

function DODefDataTypeFromDBFieldType(AFieldType: TFieldType): TibeDODefDataType;
function DODefDataTypeFromString(const ADataType: string): TibeDODefDataType;

type

TibeDODefAggFunc = (doafNone, doafCount, doafSum, doafAvg, doafMin, doafMax);

const DODefAggFuncIDArray: array[TibeDODefAggFunc] of string = ('', 'count', 'sum', 'avg', 'min', 'max');

function DODefAggFuncTypeFromString(const AAggFuncID: string): TibeDODefAggFunc;

type

TibeDOPKValueMode = (pkmCustom,     // die konkrete DO-Klasse generiert den PK
                     pkmUIDBin,     // 16 Byte Hex Zahl generiert vom DataObject
                     pkmUIDString,  // UID als 32 Zeichen Hex-String, generiert vom DataObject
                     pkmSerial,     // vom DBMS gerneriert und vom DataObject ausgelesen
                     pkmNone);

const PKValueModeArray: array[TibeDOPKValueMode] of string = ('CUSTM', 'UIDBIN', 'UIDSTR', 'SERIL', '');
const PKValueModeNameArray: array[TibeDOPKValueMode] of string = ('DO spezifisch', 'UID Hexzahl', 'UID-String', 'Serial', 'Ohne');

function PKValueModeTypeFromString(const APKValueModePKValueMode: string): TibeDOPKValueMode;

type

TibeDOTableJoinType = (dojtMasterTable, dojtJoin, dojtLeftJoin);

const TableJoinTypeArray: array[TibeDOTableJoinType] of string = ('', 'join', 'left join');

type

EibeDataObject = class(EibeException);
EibeDOFetchCountMismatch = class(EibeDataObject);

implementation

function DODefOrderByTypeFromString(const AOrderByID: string): TibeOrderByDirection;
begin
  if SameText(AOrderByID, DODefOrderByIDArray[obdAscending]) then
  begin
    Result := obdAscending;
  end else if SameText(AOrderByID, DODefOrderByIDArray[obdDescending]) then
  begin
    Result := obdDescending;
  end else
  begin
    raise EibeDataObject.Create('Nicht untersttze Sortierrichtung "' + AOrderByID + '"');
  end;
end;

function DODefDataTypeFromDBFieldType(AFieldType: TFieldType): TibeDODefDataType;
begin
  case AFieldType of
    ftString,
    ftFixedChar,
    ftWideString,
    ftFixedWideChar: Result := dodtString;
    ftWideMemo,
    ftMemo: Result := dodtMemo;
    ftAutoInc,
    ftSmallint,
    ftInteger,
    ftLargeint,
    ftLongWord,
    ftShortint,
    ftByte,
    ftSingle,
    ftWord: Result := dodtInteger;
    ftExtended,
    ftFloat: Result := dodFloat;
    ftCurrency: Result := dodtMoney;
    ftBoolean: Result := dodtBoolean;
    ftDate: Result := dodtDate;
    ftTime: Result := dodtTime;
    ftDateTime: Result := dodtDateTime;
  else begin
    raise Exception.Create('DODefDataTypeFromDBFieldType() FieldType ' + FieldTypeNames[AFieldType] + ' wird nicht abgebildet');
  end;
//    ftBytes: Result := dodt;
//    ftVarBytes: ;
//    ftAutoInc: ;
//    ftBlob: ;
//    ftGraphic: ;
//    ftFmtMemo: ;
//    ftParadoxOle: ;
//    ftDBaseOle: ;
//    ftTypedBinary: ;
//    ftCursor: ;
//    ftADT: ;
//    ftArray: ;
//    ftReference: ;
//    ftDataSet: ;
//    ftOraBlob: ;
//    ftOraClob: ;
//    ftVariant: ;
//    ftInterface: ;
//    ftIDispatch: ;
//    ftGuid: ;
//    ftTimeStamp: ;
//    ftFMTBcd: ;
//    ftOraTimeStamp: ;
//    ftOraInterval: ;
//
//    ftConnection: ;
//    ftParams: ;
//    ftStream: ;
//    ftTimeStampOffset: ;
//    ftObject: ;
  end;
end;

function DODefDataTypeFromString(const ADataType: string): TibeDODefDataType;
var
  LDataType: TibeDODefDataType;
begin
  Result := dodtString;

  for LDataType := Low(TibeDODefDataType) to High(TibeDODefDataType) do
  begin
    if SameText(ADataType, DODefDataTypeIDArray[LDataType]) then
    begin
      Result := LDataType;
      Break;
    end;
  end;
end;

function DODefAggFuncTypeFromString(const AAggFuncID: string): TibeDODefAggFunc;
var
  LFound: Boolean;
  LFunc: TibeDODefAggFunc;
begin
  Result := doafNone;
  LFound := False;

  for LFunc := Low(TibeDODefAggFunc) to High(TibeDODefAggFunc) do
  begin
    if SameText(AAggFuncID, DODefAggFuncIDArray[LFunc]) then
    begin
      Result := LFunc;
      LFound := True;
      Break;
    end;
  end;

  if not LFound then
  begin
    raise Exception.Create('DODefAggFuncTypeFromString(): nicht untersttzte Aggregatfunktion "' + AAggFuncID + '"');
  end;
end;

function PKValueModeTypeFromString(const APKValueModePKValueMode: string): TibeDOPKValueMode;
var
  LPKValueMode: TibeDOPKValueMode;
begin
  Result := pkmNone;
  
  for LPKValueMode := Low(TibeDOPKValueMode) to High(TibeDOPKValueMode) do
  begin
    if SameText(APKValueModePKValueMode, PKValueModeArray[LPKValueMode]) then
    begin
      Result := LPKValueMode;
      Break;
    end;
  end;
end;                                                             

end.
