// ============================================================================
//
// Unit:    ibeFuncCommon
// Inhalt:  Funktionen und Konstanten
// Aufgabe: Funktionsbibliothek
//
//   (c) by Klaus Etscheid since 2015
// ============================================================================
unit ibeFuncCommon;

interface

uses
  SysUtils, Math,
  ibeTypes;

Const
	ZERO	=  0.0;
	ONE		=  1.0;
	TWO		=  2.0;
	THREE	=  3.0;
	FOUR	=  4.0;
	FIVE	=  5.0;
	SIX		=  6.0;
	SEVEN	=  7.0;
	EIGHT	=  8.0;
	NINE	=  9.0;
	TEN		= 10.0;
	HALF	=  0.5;

procedure ibeFreeAndNil(var Obj);

// Methoden fr Fliekomma-Typen
Function ibeMachEps: ibeFloat;

function ibeValueIsValidNumber(val: ibeFloat): Boolean; inline;
Function ibeValueIsZero( val: ibeFloat ): Boolean; Overload;
Function ibeValueIsZero( val, eps: ibeFloat ): Boolean; Overload;
Function ibeValueIsNotZero( val: ibeFloat ): Boolean; Overload;
Function ibeValueIsNotZero( val, eps: ibeFloat ): Boolean; Overload;
Function ibeValuesAreEqual( val1, val2: ibeFloat ): Boolean; Overload;
Function ibeValuesAreEqual( val1, val2, eps: ibeFloat ): Boolean; Overload;
Function ibeValuesAreNotEqual( val1, val2: ibeFloat ): Boolean; Overload;
Function ibeValuesAreNotEqual( val1, val2, eps: ibeFloat ): Boolean; Overload;
Function ibeValue1GTValue2( val1, val2: ibeFloat ): Boolean; Overload;
Function ibeValue1GTValue2( val1, val2, eps: ibeFloat ): Boolean; Overload;
Function ibeValue1GEValue2( val1, val2: ibeFloat ): Boolean; Overload;
Function ibeValue1GEValue2( val1, val2, eps: ibeFloat ): Boolean; Overload;

Function ibeExp10(x: ibeFloat): ibeFloat;
Function ibeLog10(x: ibeFloat): ibeFloat;
Function ibeIPower2(Exponent: Cardinal): Cardinal;
Function ibeILog10(x: ibeFloat): Integer;
Function ibeRoot(Basis, Exponent: ibeFloat): ibeFloat;
Function ibeIExp10(N: Integer): Cardinal; // 10 hoch n; fr n z.Z. nur 0..9 zulssig

function ibeRandom64(): UInt64;

implementation

Var
	m_fEpsMin: ibeFloat = 10.0;

procedure ibeFreeAndNil(var Obj);
var temp: TObject;
begin
	temp := TObject(Obj);
	if Assigned(temp) then begin
		Pointer(Obj) := nil;
		temp.Free;
	end;
end;

Function ibeMachEps: ibeFloat;
Begin
	If m_fEpsMin > ONE Then Begin
		m_fEpsMin := ONE;
		While (m_fEpsMin + ONE) > ONE Do Begin
			m_fEpsMin := m_fEpsMin * HALF;
		End;
		m_fEpsMin := m_fEpsMin * TWO;
	End;
	Result := m_fEpsMin;
End;

function ibeValueIsValidNumber(val: ibeFloat): Boolean;
begin
  Result := not (IsInfinite(val) or IsNan(val));
end;

Function ibeValueIsZero( val: ibeFloat ): Boolean;
Begin
	Result	:= ibeValueIsZero( val, ibeMachEps );
End;

Function ibeValueIsZero( val, eps: ibeFloat ): Boolean; Overload;
Begin
	Result	:= Abs( val ) <= eps;
End;

Function ibeValueIsNotZero( val: ibeFloat ): Boolean;
Begin
	Result	:= Not ibeValueIsZero( val );
End;

Function ibeValueIsNotZero( val, eps: ibeFloat ): Boolean; Overload;
Begin
	Result	:= Not ibeValueIsZero( val, eps );
End;

Function ibeValuesAreEqual( val1, val2: ibeFloat ): Boolean;
Begin
	Result	:= ibeValueIsZero( val1 - val2 );
End;

Function ibeValuesAreEqual( val1, val2, eps: ibeFloat ): Boolean;
Begin
	Result	:= ibeValueIsZero( val1 - val2, eps );
End;

Function ibeValuesAreNotEqual( val1, val2: ibeFloat ): Boolean;
Begin
	Result	:= Not ibeValuesAreEqual( val1, val2 );
End;

Function ibeValuesAreNotEqual( val1, val2, eps: ibeFloat ): Boolean;
Begin
	Result	:= Not ibeValuesAreEqual( val1, val2, eps );
End;

Function ibeValue1GTValue2( val1, val2: ibeFloat ): Boolean;
Begin
	Result := ibeValue1GTValue2(val1, val2, ibeMachEps);
End;

Function ibeValue1GTValue2( val1, val2, eps: ibeFloat ): Boolean;
Begin
	Result := (val1 - eps) >= val2;
End;

Function ibeValue1GEValue2( val1, val2: ibeFloat ): Boolean;
begin
	Result := ibeValue1GEValue2(val1, val2, ibeMachEps);
end;

Function ibeValue1GEValue2( val1, val2, eps: ibeFloat ): Boolean;
begin
	Result := (val1 + eps) >= val2;
end;

{ berechnet 10 hoch x }
Function ibeExp10(x: ibeFloat): ibeFloat;
begin
	Result := Exp(x * ln(10.0));
end;

{ berechnet den Logarithmus zur Basis 10 }
Function ibeLog10(x: ibeFloat): ibeFloat;
begin
	Result := ln(x) / ln(10.0);
end;

Function ibeIPower2(Exponent: Cardinal): Cardinal;
var msg: string;
begin
	if Exponent > 31 then begin
		msg := 'IPower2 wurde mit Exponent=IntToStr(Exponent) aufgerufen. Der Wertebereich von Exponent ist [0..31].';
		Raise Exception.Create(msg);
	end;
	Result := 1 Shl Exponent;
end;

{ liefert die Ordnung einer Zahl x. Beispiele:    }
{  -1 fr Zahlen im Intervall   0.1 =< x <    1.0 }
{   0 fr Zahlen im Intervall   1.0 =< x <   10.0 }
{   1 fr Zahlen im Intervall  10.0 =< x <  100.0 }
{   2 fr Zahlen im Intervall 100.0 =< x < 1000.0 }
Function ibeILog10(x: ibeFloat): Integer;
begin
	x := Log10(x);
	Result := Trunc(x) - Ord(Frac(x) < 0.0);
end;

Function ibeRoot(Basis, Exponent: ibeFloat): ibeFloat;
Begin
	If Exponent <> 0.0 Then
		Result := Power(Basis, 1.0 / Exponent)
	Else
		Raise Exception.Create('Die 0-te Wurzel einer Zahl ist nicht definiert...');
End;

Function ibeIExp10( N: Integer ): Cardinal; // 10 hoch n; fr n z.Z. nur 0..9 zulssig
Var I: Integer;
Begin
	If N < 0 Then Raise Exception.Create('IExp10: Negative Exponenten nicht zulssig');
	If N = 0 Then Result := 1
	Else
	begin
		Result := 10;
		If N > 1 Then For I := 2 To N Do Result := Result * 10;
	end;
End;

function ibeRandom64(): UInt64;
var
  Overlay: packed record
    a, b: Int32;
  end absolute Result;
begin
  // Falls zwei 32-Bit Integer mal nicht genau einen 64-Bit Integer ergeben einen
  // Assert auslsen
  Assert(SizeOf(Overlay) = SizeOf(Result));

  Overlay.a := Random(MaxInt);
  Overlay.b := Random(MaxInt);
end;

end.

