// ============================================================================
//
// Unit:    ibeFuncDateUtil
// Inhalt:  Formular
// Aufgabe: Funktionsbibliothek
//
//   (c) by Klaus Etscheid since 2015
// ============================================================================
unit ibeFuncDateUtil;

interface

uses
  SysUtils, StrUtils, DateUtils, Math,
  ibeString;

function IntDateToDate(aIntDate: Integer): TDateTime;
function DateToIntDate(aDate: TDateTime): Integer;

function DateToIntTime(aDate: TDateTime): Integer;
function IntTimeToDate(aIntTime: Integer): TDateTime;
procedure DecodeIntTime(aIntTime: Integer; out aHour, aMin, aSec: Integer);
function EncodeIntTime(aHour, aMin, aSec: integer): Integer;
function EncodeIntTimeAsDateTime(aHour, aMin, aSec: integer): TDateTime;

// -+-+-+- Verschiedene Formatierungen -+-+-+- 

// FormatHoursAsHHMMString() formatiert die in hours bergebenen Stunden in einen
// "hh:mm" String. Stunden werden mit vorangestellten '0'-en auf mindestens
// HourDigitCount Stellen aufgefllt.
function FormatHoursAsHHMMString(hours: Double; HourDigitCount: Integer = 2): string;
function FormatDTAsHHMMString(dt: TDateTime; HourDigitCount: Integer = 2): string;
function FormatHoursAsHHMMSSString(hours: Double; HourDigitCount: Integer = 2): string;
function FormatDTAsHHMMSSMSString(dt: TDateTime; HourDigitCount: Integer): string;
function FormatHoursAsHHMMSSMSString(hours: Double; HourDigitCount: Integer): string;

// ConvertHHMMStringTo...() - Umwandlung Zeitangabe als String im Format 'hh:mm:ss.ms'.
// Die Bestandteile hh, mm, ss drfen geliebig viele Stellen haben, die Bestandteile
// mm, ss und ms sind optional.
// - ConvertHHMMStringToHours(): in Stunden mit Dezimalanteil.
// - ConvertHHMMStringToDT(): in DateTime
function ConvertHHMMStringToHours(const hhmm: string): Extended;
function ConvertHHMMStringToDT(const hhmm: string): TDateTime;


// Feiertage in Deutschland
type
  THolyday = (hdNeujahr, hdKarfreitag, hdOsterSonntag, hdOsterMontag, hdTagDerArbeit, 
              hdHimmelfahrt, hdPfingstSonntag, hdPfingstMontag, hdFrohnleichnam,
              hdTagDerDeutschenEinheit, hdAllerheiligen, hdErsterWeihnachstag,
              hdZweiterWeihnachstag,
              hdNone);

  THolydayInYear = array [THolyday] of TDateTime;

const
  HolydayNames: array [THolyday] of string = ('Neujahr', 'Karfreitag', 'Ostersonntag', 
                'Ostermontag', 'Tag der Arbeit',  'Himmelfahrt', 'Pfingstsonntag',
                'Pfingstmontag', 'Frohnleichnam', 'Tag der deutschen Einheit', 
                'Allerheiligen', '1. Weihnachstag', '2. Weihnachstag', 'Kein Feiertag');

/// <summary>
///  Gausche Osterformel (Berechnung des Ostersonntages)
///  nach Carl Friedrich Gau
///
///  Der 22. Mrz is der frheste Termin, der 25. April der letzte, auf den Ostern fallen kann.
///  Gilt fr die Jahre 1583 (Beginn des Gregorianischen Kalenders) bis 3900.
///
///  Umsetzung: Balu der Baer
///  Quelle: http://www.delphipraxis.net/38409-ostersonntag-eines-jahres-berechnen.html
///
/// </summary>
function GetEasterSunday(AYear: Integer): TDateTime;
/// <summary>
///  Die Methode fllt das Array AHolydaysInYear mit den deutschen Feiertagen im
///  bergebenen Jahr
/// </summary>
procedure GetHolydaysInYear(AYear: Integer; var AHolydaysInYear: THolydayInYear);
function GetHolydayType(ADate: TDateTime): THolyday; overload;
function GetHolydayType(ADate: TDateTime; const AHolydaysInYear: THolydayInYear): THolyday; overload;
/// <summary>
///  Die Funktion liefert True, wenn das bergebene Datum in Deutschland, NRW
///  ein gesetzlicher Feiertag ist
/// </summary>
function GetIsHolyday(ADate: TDateTime): Boolean; overload;
/// <summary>
///  Die Funktion liefert True, wenn das bergebene Datum in Deutschland, NRW
///  ein gesetzlicher Feiertag ist
/// </summary>
function GetIsHolyday(ADate: TDateTime; const AHolydaysInYear: THolydayInYear): Boolean; overload;
  
implementation

function IntDateToDate(aIntDate: Integer): TDateTime;
var iDay, iMonth, iYear: Word;
begin
  iYear := aIntDate Div 10000;
  Dec(aIntDate, iYear * 10000);

  iMonth := aIntDate Div 100;
  Dec(aIntDate, iMonth * 100);

  iDay := aIntDate;

  Result := EncodeDate(iYear, iMonth, iDay);
end;

function DateToIntDate(aDate: TDateTime): Integer;
var iDay, iMonth, iYear: Word;
begin
  DecodeDate(aDate, iYear, iMonth, iDay);
  Result := iYear * 10000
          + iMonth * 100
          + iDay;
end;

function DateToIntTime(aDate: TDateTime): Integer;
var iHour, iMin, iSec, iMSec: Word;
begin
  DecodeTime(aDate, iHour, iMin, iSec, iMSec);

  Result := EncodeIntTime(iHour, iMin, iSec);
end;

function IntTimeToDate(aIntTime: Integer): TDateTime;
var iHour, iMin, iSec: Integer;
begin
  DecodeIntTime(aIntTime, iHour, iMin, iSec);

  Result := EncodeTime(iHour, iMin, iSec, 0);
end;

procedure DecodeIntTime(aIntTime: Integer; out aHour, aMin, aSec: Integer);
begin
  aHour := aIntTime Div 10000;
  Dec(aIntTime, aHour * 10000);

  aMin := aIntTime Div 100;
  Dec(aIntTime, aMin * 100);

  aSec := aIntTime;
end;

function EncodeIntTime(aHour, aMin, aSec: integer): Integer;
var
  LIntTime: Integer;
  LTotalSeconds: Integer;
  LHour: Integer;
  LMinute: Integer;
  LSecond: Integer;
begin
  if aHour >= 0 then begin
    LIntTime := aHour * 10000
           + aMin  *   100
           + aSec;
  end else begin
    LTotalSeconds := aHour * 3600 + aMin * 60 + aSec;

    LHour := LTotalSeconds div 3600;
    LMinute := (LTotalSeconds - 3600 * LHour) div 60;
    LSecond := LTotalSeconds - 3600 * LHour - 60 * LMinute;

    LIntTime := LHour   * 10000
              + LMinute *   100
              + LSecond;
  end;

  Result := LIntTime;
end;

function EncodeIntTimeAsDateTime(aHour, aMin, aSec: Integer): TDateTime;
var dt: TDateTime;
  iDays: Integer;
begin
  if aHour < 0 then begin
    iDays := (aHour + 1) div 24 - 1;
    if iDays = 0 then begin
      iDays := -1;
    end;
    Dec(aHour, iDays * 24);
    dt := IncHour(iDays, aHour);
  end else begin
    dt := IncHour(0.0, aHour);
  end;

  dt := IncMinute(dt, aMin);
  dt := IncSecond(dt, aSec);

  Result := dt;
end;

// -+-+-+- Verschiedene Formatierungen -+-+-+-

function FormatDTAsHHMMString(dt: TDateTime; HourDigitCount: Integer): string;
begin
	Result	:= FormatHoursAsHHMMString(dt * 24.0, HourDigitCount);
end;

function FormatHoursAsHHMMString(hours: Double; HourDigitCount: Integer): string;
var ze, h, m: Integer;
	sign: string;
Begin
	sign		:= IfThen(hours < 0, '-', '');
	ze			:= Round(Abs(hours) * 60.0);
	h				:= ze Div 60;
	m				:= ze Mod 60;
	Result	:= Format('%s%*.*d:%2.2d', [sign, HourDigitCount, HourDigitCount, h, m]);
End;

function FormatHoursAsHHMMSSString(hours: Double; HourDigitCount: Integer = 2): string;
var ze, h, m, s: Integer;
	seconds: Extended;
	sign: string;
begin
	sign		:= IfThen(hours < 0, '-', '');
	seconds	:= Abs(hours) * 60.0 * 60.0;
	ze			:= Round(seconds);
	h				:= ze Div (60 * 60);
	m				:= (ze - h * 60 * 60) Div 60;
	s				:= ze - 60 * 60 * h - 60 * m;
	Result	:= Format('%s%*.*d:%2.2d:%2.2d', [sign, HourDigitCount, HourDigitCount, h, m, s]);
end;

function FormatDTAsHHMMSSMSString(dt: TDateTime; HourDigitCount: Integer): string;
begin
	Result	:= FormatHoursAsHHMMSSMSString(dt * 24.0, HourDigitCount);
end;

function FormatHoursAsHHMMSSMSString(hours: Double; HourDigitCount: Integer): string;
var ze, h, m, s, ms: Integer;
	seconds: Extended;
	sign: string;
begin
	sign		:= IfThen(hours < 0, '-', '');
	seconds	:= Abs(hours) * 60.0 * 60.0;
	ze			:= Round(seconds);
	h				:= ze Div (60 * 60);
	m				:= (ze - h * 60 * 60) Div 60;
	s				:= ze - 60 * 60 * h - 60 * m;
	ms			:= Round(Frac(seconds) * 1000.0) div 1000;
	Result	:= Format('%s%*.*d:%2.2d:%2.2d.%3.3d', [sign, HourDigitCount, HourDigitCount, h, m, s, ms]);
end;

function ConvertHHMMStringToHours(const hhmm: string): Extended;
var h, m, s, ms, sign: Integer;
	sHHMM, sHHMMSS, sMS: string;
begin
	sign  := 1;
	sHHMM := Trim(hhmm);

	if sHHMM <> '' then begin
		if sHHMM[1] = '-' then begin
			sign  := -1;
			sHHMM := Copy(sHHMM, 2, Length(sHHMM));
		end;
	end;

	sHHMMSS	:= ExtractWord(1, sHHMM, ['.']);
	sMS			:= ExtractWord(2, sHHMM, ['.']);

	h	:= StrToInt(ExtractWord(1, sHHMMSS, [':']));
	m	:= StrToIntDef(ExtractWord( 2, sHHMMSS, [':']), 0);
	s	:= StrToIntDef(ExtractWord( 3, sHHMMSS, [':']), 0);

	Result := sign * (h + m / 60.0 + s / 3600.0);

	if sMS <> '' then begin
		ms := StrToInt(sMS);
		Result := Result + ms / (60.0 * 60.0 * 1000.0);
	end;
end;

function ConvertHHMMStringToDT(const hhmm: string): TDateTime;
begin
	Result := ConvertHHMMStringToHours(hhmm) / 24.0;
end;

(* Gausche Osterformel (Berechnung des Ostersonntages)

  nach Carl Friedrich Gau

  Der 22. Mrz is der frheste Termin, der 25. April der letzte, auf den Ostern fallen kann.
  Gilt fr die Jahre 1583 (Beginn des Gregorianischen Kalenders) bis 3900.

  Umsetzung: Balu der Baer
  Quelle: http://www.delphipraxis.net/38409-ostersonntag-eines-jahres-berechnen.html
*)
function GetEasterSunday(AYear: Integer): TDateTime;
var
  a, b, c, d, e,
  H1, H2, n, m,
  temp, easter : Integer;
begin
  if not ((AYear > 1583) and (AYear < 3900)) then
  begin
    raise Exception.CreateFmt('function EasterSunday: parameter Year must be between 1583 and 3900 but is %d', [AYear])
  end else
  begin
    a := AYear mod 19;
    b := AYear mod 4;
    c := AYear mod 7;
    H1 := AYear div 100;
    H2 := AYear div 400;
    n := 4 + H1 - H2;
    m := 15 + H1 - H2 - ((8 * H1 + 13) div 25);
    d := (19 * a + m) mod 30;
    e := (2 * b + 4 * c + 6 * d + n) mod 7;

//  die nachfolgende Zuweisung auf easter wird niemals benutzt, da der nachfolgende
//  If-Block die Variable in jedem Fall ndert, ohne dass diese zuvor benutzt
//  wurde. Den Code deshalb auskommentiert.
//    if d + e = 35 then
//    begin
//      easter := 50;
//    end;
    
    if (d = 28) and (e = 6) and (a > 10) then
    begin
      easter := 49
    end else
    begin
      easter := 22 + d + e;
    end;

    if easter > 31 then
    begin
      temp := easter - 31;
      case temp of
        26 : temp := 19;
        25 : if (d = 28) and (a > 10) then temp := 18;
      end;
      Result := EncodeDate(AYear, 4, temp);
    end else
    begin
      Result := EncodeDate(AYear, 3, easter);
    end;
  end
end;

procedure GetHolydaysInYear(AYear: Integer; var AHolydaysInYear: THolydayInYear);
begin
  AHolydaysInYear[hdNone] := 0;

  // statische Feiertage
  AHolydaysInYear[hdNeujahr] := EncodeDate(AYear, 1, 1);
  AHolydaysInYear[hdTagDerArbeit] := EncodeDate(AYear, 5, 1);
  AHolydaysInYear[hdTagDerDeutschenEinheit] := EncodeDate(AYear, 10, 3);
  AHolydaysInYear[hdErsterWeihnachstag] := EncodeDate(AYear, 12, 25);
  AHolydaysInYear[hdZweiterWeihnachstag] := EncodeDate(AYear, 12, 26);
  AHolydaysInYear[hdAllerheiligen] := EncodeDate(AYear, 11, 1);

  // von Ostern abhngige Feiertage
  AHolydaysInYear[hdOsterSonntag] := GetEasterSunday(AYear);
  AHolydaysInYear[hdKarfreitag] := AHolydaysInYear[hdOsterSonntag] - 2;
  AHolydaysInYear[hdOsterMontag] := AHolydaysInYear[hdOsterSonntag] + 1;
  AHolydaysInYear[hdHimmelfahrt] := AHolydaysInYear[hdOsterSonntag] + 39;
  AHolydaysInYear[hdPfingstSonntag] := AHolydaysInYear[hdOsterSonntag] + 49;
  AHolydaysInYear[hdPfingstMontag] := AHolydaysInYear[hdPfingstSonntag] + 1;
  AHolydaysInYear[hdFrohnleichnam] := AHolydaysInYear[hdOsterSonntag] + 60;
end;

function GetHolydayType(ADate: TDateTime): THolyday; overload;
var
  LHolydaysInYear: THolydayInYear;
begin
  GetHolydaysInYear(YearOf(ADate), LHolydaysInYear);

  Result := GetHolydayType(ADate, LHolydaysInYear);
end;

function GetHolydayType(ADate: TDateTime; const AHolydaysInYear: THolydayInYear): THolyday; overload;
var
  LHolyday: THolyday;
begin
  Result := hdNone;
  
  for LHolyday := Low(THolyday) to Pred(hdNone) do
  begin
    if ADate = AHolydaysInYear[LHolyday] then
    begin
      Result := LHolyday;
      Break;
    end;
  end;
end;

function GetIsHolyday(ADate: TDateTime): Boolean; overload;
var
  LHolydaysInYear: THolydayInYear;
begin
  GetHolydaysInYear(YearOf(ADate), LHolydaysInYear);

  Result := GetIsHolyday(ADate, LHolydaysInYear);
end;

function GetIsHolyday(ADate: TDateTime; const AHolydaysInYear: THolydayInYear): Boolean; overload;
var
  LHolyday: THolyday;
begin
  Result := False;
  
  for LHolyday := Low(THolyday) to Pred(hdNone) do
  begin
    if ADate = AHolydaysInYear[LHolyday] then
    begin
      Result := True;
      Break;
    end;
  end;
end;

end.
