unit ibeProgramFunctionManager;

interface

uses
  System.SysUtils, System.StrUtils, System.Classes, Generics.Collections, Generics.Defaults,
  ibeFCommon;

type

TibeProgramFunctionType = (pfList, pfContainer, pfEdit, pfNone);

const
  ProgramFunctionTypeNames: array[TibeProgramFunctionType] of string = ('List', 'Container', 'Edit', 'None');

procedure RegisterProgramFunction(const AProgramFunctionID: string; AProgramFunctionType: TibeProgramFunctionType; AProgramFunctionFormClass: TfrmCommonClass);
function ProgramFunctionExists(const AProgramFunctionID: string; AProgramFunctionType: TibeProgramFunctionType): Boolean;
function CreateProgramFunction(AOwner: TComponent; const AConnectionName, AProgramFunctionID: string; AProgramFunctionType: TibeProgramFunctionType): TfrmCommon; overload;

implementation

// ----------------- [ TibeProgramFunctionManager ] ----------------------------

type

TibeProgramFunction = class
private
  FProgramFunctionID: string;
  FProgramFunctionType: TibeProgramFunctionType;
  FProgramFunctionFormClass: TfrmCommonClass;
protected
public
  constructor Create(const AProgramFunctionID: string; AProgramFunctionType: TibeProgramFunctionType; AProgramFunctionFormClass: TfrmCommonClass); virtual;
  function Compare(AProgramFunction: TibeProgramFunction): Integer;
  property ProgramFunctionID: string read FProgramFunctionID;
  property ProgramFunctionType: TibeProgramFunctionType read FProgramFunctionType;
  property ProgramFunctionFormClass: TfrmCommonClass read FProgramFunctionFormClass;
end;

TibeProgramFunctionList = class(TObjectList<TibeProgramFunction>)
private
  FIsSorted: Boolean;
protected
public
  constructor Create(AComparer: TComparison<TibeProgramFunction>);
  function Add(const AProgramFunction: TibeProgramFunction): Integer;
  procedure Sort();
  property IsSorted: Boolean read FIsSorted;
end;

TibeProgramFunctionManager = class
private
  FProgramFunctions: TibeProgramFunctionList;
protected
  property ProgramFunctions: TibeProgramFunctionList read FProgramFunctions;
public
  constructor Create();
  destructor Destroy(); override;
  procedure RegisterProgramFunction(const AProgramFunctionID: string; AProgramFunctionType: TibeProgramFunctionType; AProgramFunctionFormClass: TfrmCommonClass);
  function ProgramFunctionExists(const AProgramFunctionID: string; AProgramFunctionType: TibeProgramFunctionType): Boolean;
  function CreateProgramFunction(AOwner: TComponent; const AConnectionName, AProgramFunctionID: string; AProgramFunctionType: TibeProgramFunctionType): TfrmCommon; overload;
end;

constructor TibeProgramFunctionManager.Create();
var
  LProgFuncComparer: TComparison<TibeProgramFunction>;
begin
  inherited Create();

  LProgFuncComparer := function(const Left, Right: TibeProgramFunction): Integer
                       begin
                         Result := Left.Compare(Right);
                       end;

  FProgramFunctions := TibeProgramFunctionList.Create(LProgFuncComparer);
end;

destructor TibeProgramFunctionManager.Destroy();
begin
  inherited Destroy();

  FreeAndNil(FProgramFunctions);
end;

procedure TibeProgramFunctionManager.RegisterProgramFunction(const AProgramFunctionID: string; AProgramFunctionType: TibeProgramFunctionType; AProgramFunctionFormClass: TfrmCommonClass);
var
  LProgramFunction: TibeProgramFunction;
begin
  LProgramFunction := TibeProgramFunction.Create(AProgramFunctionID, AProgramFunctionType, AProgramFunctionFormClass);
  try
    ProgramFunctions.Add(LProgramFunction);
  except
    FreeAndNil(LProgramFunction);
  end;
end;

function TibeProgramFunctionManager.ProgramFunctionExists(const AProgramFunctionID: string; AProgramFunctionType: TibeProgramFunctionType): Boolean;
var
  LIndex: Integer;
  LProgramFunction: TibeProgramFunction;
begin
  if not ProgramFunctions.IsSorted then
  begin
    ProgramFunctions.Sort();
  end;

  LProgramFunction := TibeProgramFunction.Create(AProgramFunctionID, AProgramFunctionType, nil);
  try
    Result := ProgramFunctions.BinarySearch(LProgramFunction, LIndex);
  finally
    FreeAndNil(LProgramFunction);
  end;
end;

function TibeProgramFunctionManager.CreateProgramFunction(AOwner: TComponent; const AConnectionName, AProgramFunctionID: string; AProgramFunctionType: TibeProgramFunctionType): TfrmCommon;
var
  LIndex: Integer;
  LProgramFunction: TibeProgramFunction;
  LProgramFunctionForm: TfrmCommon;
const
  LMessageFunctionNotFound = '%s.CreateProgramFunction(): Programmfunktion "%s" ist fr die Funktionsart %s nicht registiert';
begin
  LProgramFunctionForm := nil;

  if not ProgramFunctions.IsSorted then
  begin
    ProgramFunctions.Sort();
  end;

  LProgramFunction := TibeProgramFunction.Create(AProgramFunctionID, AProgramFunctionType, nil);
  try
    if ProgramFunctions.BinarySearch(LProgramFunction, LIndex) then
    begin
      LProgramFunctionForm := ProgramFunctions[LIndex].ProgramFunctionFormClass.Create(AOwner);
    end else
    begin
      raise Exception.CreateFmt(LMessageFunctionNotFound, [ClassName, AProgramFunctionID, ProgramFunctionTypeNames[AProgramFunctionType]]);
    end;
  finally
    FreeAndNil(LProgramFunction);
  end;

  Result := LProgramFunctionForm;
end;

// ------------------------- [ TibeProgramFunction ] ---------------------------

constructor TibeProgramFunctionList.Create(AComparer: TComparison<TibeProgramFunction>);
begin
  inherited Create(TComparer<TibeProgramFunction>.Construct(AComparer));

  FIsSorted := True;
end;

function TibeProgramFunctionList.Add(const AProgramFunction: TibeProgramFunction): Integer;
begin
  Result := inherited Add(AProgramFunction);

  FIsSorted := False;
end;

procedure TibeProgramFunctionList.Sort();
begin
  inherited Sort();

  FIsSorted := True;
end;

constructor TibeProgramFunction.Create(const AProgramFunctionID: string; AProgramFunctionType: TibeProgramFunctionType; AProgramFunctionFormClass: TfrmCommonClass);
begin
  inherited Create();
  FProgramFunctionID := AProgramFunctionID;
  FProgramFunctionType := AProgramFunctionType;
  FProgramFunctionFormClass := AProgramFunctionFormClass;
end;

function TibeProgramFunction.Compare(AProgramFunction: TibeProgramFunction): Integer;
begin
  Result := CompareText(ProgramFunctionID, AProgramFunction.ProgramFunctionID);
  if Result = 0 then
  begin
    Result := Ord(ProgramFunctionType) - Ord(AProgramFunction.ProgramFunctionType);
  end;
end;

var
  MG_ProgramFunctionManager: TibeProgramFunctionManager;

procedure RegisterProgramFunction(const AProgramFunctionID: string; AProgramFunctionType: TibeProgramFunctionType; AProgramFunctionFormClass: TfrmCommonClass);
begin
  MG_ProgramFunctionManager.RegisterProgramFunction(AProgramFunctionID, AProgramFunctionType, AProgramFunctionFormClass);
end;

function ProgramFunctionExists(const AProgramFunctionID: string; AProgramFunctionType: TibeProgramFunctionType): Boolean;
begin
  Result := MG_ProgramFunctionManager.ProgramFunctionExists(AProgramFunctionID, AProgramFunctionType);
end;

function CreateProgramFunction(AOwner: TComponent; const AConnectionName, AProgramFunctionID: string; AProgramFunctionType: TibeProgramFunctionType): TfrmCommon; overload;
begin
  Result := MG_ProgramFunctionManager.CreateProgramFunction(AOwner, AConnectionName, AProgramFunctionID, AProgramFunctionType);
end;

initialization

  MG_ProgramFunctionManager := TibeProgramFunctionManager.Create();

finalization

  FreeAndNil(MG_ProgramFunctionManager);

end.

