// ============================================================================
//
// Unit:    ibeXMLDocument
// Inhalt:  Formular
// Aufgabe: Klassenbibliothek
//
//   (c) by Klaus Etscheid since 2015
// ============================================================================
unit ibeXMLDocument;

interface

uses
  SysUtils, System.Classes,
  XMLDoc, XMLIntf, XMLDOM, WideStrUtils,
  ibeString, ibeVariant, ibeTypes;

type

IXMLNode = XMLIntf.IXMLNode;

TibeXMLBaseDocument = class
private
	FxmlDoc: IXMLDocument;
	Function GetDocumentElement: IXMLNode;
	Function GetxmlText: DOMString;
	Procedure SetxmlText(const xml: DOMString);
protected
	Property xmlDoc: IXMLDocument Read FxmlDoc;
	Property xmlText: DOMString Read GetxmlText Write SetxmlText;
public
	Constructor Create;
	Destructor Destroy; Override;
	Property DocumentElement: IXMLNode Read GetDocumentElement;
	Procedure CreateDocument(Const DocTypeName: DOMString; Const NameSpace: DOMString = '');
	Procedure LoadFromFile(const a_sFileName: TFileName);
	Procedure LoadFromStream(AStream: TStream);
	Procedure SaveToFile(const a_sFileName: TFileName);
	Procedure SaveToStream(AStream: TStream);
end;

TibeXMLBaseElement = class
Private
	FLastNodeIndex: Integer;
	FCurrentNode, FLastParentNode: IXMLNode;
	FAllowMissingNodes: Boolean;
	FPathDelimeter: Char;
	Function GetParentNode(a_ParentNode: IXMLNode = nil): IXMLNode;
	Procedure HandleNodeNotFound(a_ParentNode: IXMLNode; const NodeName, NodeType: string);
Protected
	Property CurrentNode: IXMLNode Read FCurrentNode Write FCurrentNode;
	Property LastNodeIndex: Integer Read FLastNodeIndex;
	Property LastParentNode: IXMLNode Read FLastParentNode;
	Property AllowMissingNodes: Boolean Read FAllowMissingNodes Write FAllowMissingNodes;
	Property PathDelimeter: Char Read FPathDelimeter;
Public
	Constructor Create();
	Function GetNodeByName(const PathAndName: DOMString; ParentNode: IXMLNode = nil): IXMLNode;
	Function GetNextChildNode(const NodeName: DOMString; ParentNode: IXMLNode = nil; FromFirstNode: Boolean = False): IXMLNode;
	Function GetAttribute(const Name: DOMString; ParentNode: IXMLNode = nil): IXMLNode;
	Function GetAttributeTypedValue(const Name: DOMString; Default: Variant; ParentNode: IXMLNode = nil): Variant;
	Function GetAttributeValue(const Name: DOMString; Default: Variant; ParentNode: IXMLNode = nil): Variant;
	Function CreateElement(const Name: DOMString; ParentNode: IXMLNode = nil): IXMLNode;
	Function CreateAttribute(const Name: DOMString; Value: Variant; ParentNode: IXMLNode = nil): IXMLNode;
End;

TibeXMLDocument = Class(TibeXMLBaseDocument)
Private
Protected
Public
	Property xmlText;
End;

TibeXMLElement = Class(TibeXMLBaseElement)
Private
Protected
Public
	Property CurrentNode;
	Property AllowMissingNodes;
End;

const
	XMLPathDelimeter	= '/';

Function EvaluateNodePath( node: IXMLNode; PathDelimeter: Char = XMLPathDelimeter ): String;

implementation

Constructor TibeXMLBaseDocument.Create;
Begin
	Inherited Create;

	FxmlDoc := TXMLDocument.Create(nil);
	xmlDoc.Options := xmlDoc.Options + [doNodeAutoIndent];
End;

Destructor TibeXMLBaseDocument.Destroy;
Begin
	Inherited Destroy;

	FxmlDoc := nil;
End;

Procedure TibeXMLBaseDocument.CreateDocument(const DocTypeName: DOMString; Const NameSpace: DOMString);
Begin
	xmlDoc.Active := True;
	xmlDoc.Encoding := 'UTF-8';
	xmlDoc.Version := '1.0';
	xmlDoc.DocumentElement := xmlDoc.CreateElement(DocTypeName, NameSpace);
End;

Procedure TibeXMLBaseDocument.LoadFromFile(const a_sFileName: TFileName);
Begin
	xmlDoc.LoadFromFile(a_sFileName);
End;

Procedure TibeXMLBaseDocument.LoadFromStream(AStream: TStream);
Begin
	xmlDoc.LoadFromStream(AStream);
End;

Procedure TibeXMLBaseDocument.SaveToFile(const a_sFileName: TFileName);
Begin
	xmlDoc.SaveToFile(a_sFileName);
End;

Procedure TibeXMLBaseDocument.SaveToStream(AStream: TStream);
Begin
	xmlDoc.SaveToStream(AStream);
End;

Function TibeXMLBaseDocument.GetDocumentElement: IXMLNode;
Begin
	Result := xmlDoc.DocumentElement;
End;

Function TibeXMLBaseDocument.GetxmlText: DOMString;
Begin
	If Assigned(xmlDoc.XML) Then Result := xmlDoc.XML.Text
	Else										 		 Result := '';
End;

Procedure TibeXMLBaseDocument.SetxmlText(const xml: DOMString);
Begin
	If xmlDoc.Active Then xmlDoc.Active := False;
	xmlDoc.LoadFromXML(xml);
	xmlDoc.Active	:= True;
End;

{ -+-+-+- TibeXMLBaseNode -+-+-+- }

Constructor TibeXMLBaseElement.Create();
Begin
	Inherited Create;
	FLastParentNode := Nil;
	FLastNodeIndex := -1;
	AllowMissingNodes := True;
	FPathDelimeter := XMLPathDelimeter;
End;

Function TibeXMLBaseElement.GetNodeByName(const PathAndName: DOMString; ParentNode: IXMLNode = nil): IXMLNode;
Var n: IXMLNode;
	i, j, NodeCount: Integer;
	NodeName: DOMString;
	ChildNodeFound: Boolean;
Begin
	n := GetParentNode( ParentNode );
	ChildNodeFound := False;

	Result := Nil;
	NodeCount := WordCount( PathAndName, [PathDelimeter] );
	For i := 1 To NodeCount Do Begin
		NodeName := ExtractWord( i, PathAndName, [PathDelimeter] );
	{ Wenn der letzte Node in FolderAndName gesucht wird, dann zunchst in der }
	{ Attributliste suchen.                                                    }
		If i = NodeCount Then Begin
			Result := n.AttributeNodes.FindNode( NodeName );
		End;
		If Not Assigned( Result ) Then Begin
		{ Entweder ist FolderAndName noch nicht abgesucht, oder es wurde kein Attribut }
		{ mit passendem Namen gefunden. Jetzt die ChildNodes durchsuchen }
			For j := 0 To n.ChildNodes.Count - 1 Do Begin
				If n.ChildNodes[j].NodeName = NodeName Then Begin
					ChildNodeFound := True;
					n := n.ChildNodes[j];
					Break;
				End;
			End;
			If Not ChildNodeFound Then HandleNodeNotFound( GetParentNode( ParentNode ), PathAndName, 'Element oder Atribut' );
		End;
		If (i = NodeCount) And ChildNodeFound Then Begin
			Result := n;
		End;
	End;
End;

Function TibeXMLBaseElement.GetNextChildNode(const NodeName: DOMString; ParentNode: IXMLNode = nil;
	FromFirstNode: Boolean = False): IXMLNode;
Var n: IXMLNode;
	i: Integer;
Begin
	n := GetParentNode( ParentNode );

	If FromFirstNode Or (n <> LastParentNode) Then Begin
		FLastNodeIndex := -1;
		FLastParentNode := n;
	End;

	Result := Nil;
	For i := LastNodeIndex + 1 To n.ChildNodes.Count - 1 Do Begin
		If n.ChildNodes[i].NodeName = NodeName Then Begin
			FLastNodeIndex := i;
			Result := n.childNodes[i];
			//CurrentNode := Result;
			Break;
		End;
	End;
End;

Function TibeXMLBaseElement.GetAttribute(const Name: DOMString; ParentNode: IXMLNode = nil): IXMLNode;
Var n: IXMLNode;
Begin
	n := GetParentNode( ParentNode );
	Result := n.AttributeNodes.FindNode( Name );
	If Not Assigned( Result ) Then HandleNodeNotFound( n, Name, 'Attribut' );
End;

Function TibeXMLBaseElement.GetAttributeTypedValue(const Name: DOMString; Default: Variant; ParentNode: IXMLNode = nil): Variant;
var node: IXMLNode;
  DataType: TibeDatatype;
begin
	Result := Default;
	node := GetAttribute( Name, ParentNode );
	If Assigned( node ) Then Begin
    DataType := ibeDatatypeFromVariant(Default);
		Result := XMLStringAsVariant(VarAsString(node.NodeValue), DataType);
		If DataType = dtString Then Result := WideStringReplace(VarAsString(Result), #10, #13#10, [rfReplaceAll]);
	End;
End;

Function TibeXMLBaseElement.GetAttributeValue(const Name: DOMString; Default: Variant; ParentNode: IXMLNode = nil): Variant;
Var node: IXMLNode;
Begin
	node := GetAttribute(Name, ParentNode);

	if Assigned(node) then 
  begin
    Result := node.NodeValue;
  end else
  begin
    Result := Default;
  end;
End;

Function TibeXMLBaseElement.CreateElement(const Name: DOMString; ParentNode: IXMLNode): IXMLNode;
Begin
  Result := GetParentNode(ParentNode).AddChild(Name);
End;

Function TibeXMLBaseElement.CreateAttribute(const Name: DOMString; Value: Variant; ParentNode: IXMLNode): IXMLNode;
Begin
  GetParentNode(ParentNode).Attributes[Name] := VarAsXMLString(Value);
	Result := GetParentNode(ParentNode).AttributeNodes[Name];
End;

Function TibeXMLBaseElement.GetParentNode(a_ParentNode: IXMLNode): IXMLNode;
Begin
	If			Assigned(a_ParentNode)							Then Result := a_ParentNode
	Else If	Assigned(CurrentNode)							  Then Result := CurrentNode
	Else Raise Exception.Create( 'TibeXMLBaseNode: Methodenaufruf nicht verfgbar, ParentNode ist nicht vorhanden.' );
End;

Procedure TibeXMLBaseElement.HandleNodeNotFound(a_ParentNode: IXMLNode; Const NodeName, NodeType: String);
Begin
	If Not AllowMissingNodes Then Begin
		Raise Exception.Create(NodeType + ' ''' + NodeName + ''' in Pfad ''' + EvaluateNodePath(a_ParentNode) + ''' wurde nicht gefunden.');
	End;
End;

{ -+-+-+- Verschiedene Methoden -+-+-+- }

function EvaluateNodePath(node: IXMLNode; PathDelimeter: Char): String;
Begin
	Result := node.NodeName;
	While Assigned( node.ParentNode ) Do Begin
		node := node.ParentNode;
		Result := node.nodeName + PathDelimeter + Result;
	End;
End;

end.
